# include "../lib/lib.h"
#ifndef PRINT
# define PRINT
#endif

void print_server_solicit_info(SERVER_SOLICIT_INFO *head) {
  SUBNET_PREFIX_INFO  *t_sp;
  char name[64];
  uint8_t prefix_len, count;

  printf("\n--------------------------------------------\n");
  for(; head; head = head->next) {
    printf("relay: %s\n", inet_ntop(AF_INET6, &head->relay_addr, name, 64));
    printf("server: %s\n", inet_ntop(AF_INET6, &head->server_addr, name, 64));
    printf("preference: %u\n", head->preference);
    printf("prefix-list: ");
    for(t_sp = head->subnet_prefix_info_head; t_sp; t_sp = t_sp->next) {
      prefix_len = t_sp->prefix_len_octets[0];
      for(count = 1; count <= (prefix_len + 7)/8; count++) {
	printf("%02x", t_sp->prefix_len_octets[count]);
	if((count % 2) == 0 && count != (prefix_len + 7)/8)
	  printf(":");
      }
      printf("    ");
    }
  }
}

void print_leased_addr_list(LIADDR *head) {
  LIADDR *t_liaddr;
  char name[64];
  uint8_t prefix_len, count;

  printf("\n----------------- leased addr list ---------------------------\n");
  for(; head; head = head->next) {
    printf("ip_addr: %s\n", inet_ntop(AF_INET6, &head->ip_addr, name, 64));
    printf("dns-name: %s\n", head->dns_name);
    printf("relay: %s\n", inet_ntop(AF_INET6, &head->relay_addr, name, 64));
    printf("server: %s\n", inet_ntop(AF_INET6, &head->server_addr, name, 64));
    printf("when-leased: %u %u\n", head->when_leased.tv_sec, head->when_leased.tv_usec);
    printf("preferred-lifetime: %u\n", head->preferred_lifetime);
    printf("valid-lifetime: %u\n\n", head->valid_lifetime);
    printf("A %d    P %d\n", head->A, head->P);
  }
  printf("\n--------------- end of leased addr list ---------------------\n");
}

SERVER_SOLICIT_INFO *ssi_head = NULL;
LIADDR *liaddr_list_head = NULL;
int fd;

char X_ADDR[] = "3ffe:305:1002:1:2b0:d0ff:fe39:6ead";

int send_and_process_release_by_extension(char *pointer, uint16_t length,
					  uint8_t r, struct in6_addr *server_addr,
					  struct in6_addr *x_addr) {
  HEADER_CLIENTD_PACKET   header;
  IP_ADDR_EXT            *iae_head = NULL, **temp_iae = &iae_head, *t_iae;
  return 1;
}

int send_and_process_init() {

  SERVER_SOLICIT_INFO **temp_ssi, *t_ssi;
  SUBNET_PREFIX_INFO  **temp_sp, *t_sp;
  uint8_t i, no_of_prefixes;
  uint32_t length_read;
  HEADER_CLIENTD_PACKET   header;

  printf("Sending init to client daemon\n");
  i = init(&fd, 1, 0, 1, 1, 1);
  if(i == 1) {
    printf("\n\nResult of solicit = Success. Continuing...\n");
    length_read = 0;
    temp_ssi    = &ssi_head;
    t_ssi       = *temp_ssi;

    READ_DATA(&header, recv_msg, length_read, sizeof(header));
    header.pid        = ntohs(header.pid);
    header.request_no = ntohl(header.request_no);
    
    while(length_read < n) {
      if(!t_ssi) {
	t_ssi = (SERVER_SOLICIT_INFO *) malloc(sizeof(SERVER_SOLICIT_INFO));
	if(!t_ssi)
	  errx(1, "Insufficient memory. Exiting...\n");
	bzero((char *) t_ssi, sizeof(*t_ssi));
      }
      READ_DATA(&t_ssi->relay_addr, recv_msg, length_read,
		sizeof(t_ssi->relay_addr));
      READ_DATA(&t_ssi->server_addr, recv_msg, length_read,
		sizeof(t_ssi->server_addr));
      /* Skip reserved field */
      length_read += 2;
      READ_DATA(&t_ssi->preference, recv_msg, length_read,
		sizeof(t_ssi->preference));
      READ_DATA(&t_ssi->no_of_prefixes, recv_msg, length_read,
		sizeof(t_ssi->no_of_prefixes));

      temp_sp = &t_ssi->subnet_prefix_info_head;
      t_sp = *temp_sp;

      for(no_of_prefixes = 0; no_of_prefixes < t_ssi->no_of_prefixes;
	  no_of_prefixes++) {
	if(!t_sp) {
	  t_sp = (SUBNET_PREFIX_INFO *) malloc(sizeof(SUBNET_PREFIX_INFO));
	  if(!t_sp)
	    errx(1, "Insufficient memory. Exiting...\n");
	  bzero((char *) t_sp, sizeof(*t_sp));
	}
	READ_DATA(t_sp->prefix_len_octets, recv_msg, length_read,
		  (((*((uint8_t *)(recv_msg + length_read))) + 7) / 8) + 1);
	*temp_sp = t_sp;
	temp_sp = &(*temp_sp)->next;
	t_sp    = t_sp->next;
      }
      
      /* free up the remaining list if present */
      while(t_sp) {
	t_sp = t_sp->next;
	free(*temp_sp);
	*temp_sp = t_sp;
      }
      *temp_sp = NULL;

      /* Consume the padding making length_read a multiple of 4 */
      length_read = ((length_read + 3) / 4) * 4;
      *temp_ssi = t_ssi;
      temp_ssi = &(*temp_ssi)->next;
      t_ssi    = t_ssi->next;
    }

    /* free up the remaining list if present */
    while(t_ssi) {
      t_ssi = t_ssi->next;
      free(*temp_ssi);
      *temp_ssi = t_ssi;
    }
    *temp_ssi = NULL;
  }
  else {
    printf("\n\nResult of solicit = Failure. Returning...\n");
    return 0;
  }

  return 1;

}

int is_present(struct in6_addr *ip_addr, struct in6_addr addr[], int length) {
  int i;

  for(i = 0; i < length; i++)
    if(IN6_ARE_ADDR_EQUAL(ip_addr, &addr[i]))
      return 1;

  return 0;
}

int send_and_process_request_clear(struct in6_addr addr[], int length) {
  HEADER_CLIENTD_PACKET   header;
  uint32_t                reserved_r_status;
  LIADDR                **temp_liaddr = &liaddr_list_head, *t_liaddr;
  IP_ADDR_EXT            *iae_head = NULL, **temp_iae = &iae_head, *t_iae;
  struct in6_addr        *req_relay_addr, *req_server_addr, rep_relay_addr;
  TYPE_LEN                tl;
  uint32_t                length_read;
  int                     i;

  bzero((char *) &rep_relay_addr, sizeof(rep_relay_addr));
  
  if(!ssi_head) {
    warnx("No server to send request to even after solicit. Returning...\n");
    return 0;
  }
  /* Build up a list of IP address extensions to be sent. These are not really
   * IP Address Extensions but a list of structures each of which can be expanded
   * into an IP Address Extension.
   */

  for(i = 0; i < length; i++) {
    t_iae = (IP_ADDR_EXT *) malloc(sizeof(IP_ADDR_EXT));
    if(!t_iae)
      errx(1, "Insufficient memory\n");

    t_iae->s.status     = 0;
    t_iae->s.C          = 1;
    t_iae->s.I          = 0;
    t_iae->s.L          = 0;
    t_iae->s.Q          = 1;
    t_iae->s.A          = 0;
    t_iae->s.P          = 0;
    t_iae->s.reserved   = 0;
    t_iae->s.scope      = 0;
    t_iae->s.prefix_len = 0;
    t_iae->ip_addr      = addr[i];
    t_iae->dns_name[0]  = '\0';

    *temp_iae = t_iae;
    temp_iae  = &(*temp_iae)->next;
  }

  req_relay_addr  = &ssi_head->relay_addr;
  req_server_addr = &ssi_head->server_addr;

  printf("Sending request to client daemon\n");
  i = request(NULL, req_relay_addr, req_server_addr, 1 /* C */, 0 /* R */,
	      0/* T */, 0/* transaction_id */, iae_head, NULL, 0, NULL, 0);
  
  if(i == 0) {
    printf("Failure in getting an address. Returning...\n");
    return 0;
  }
  length_read = 0;
  
  READ_DATA(&header, recv_msg, length_read, sizeof(header));
  header.pid        = ntohs(header.pid);
  header.request_no = ntohl(header.request_no);

  READ_DATA(&reserved_r_status, recv_msg, length_read, sizeof(reserved_r_status));
  reserved_r_status = ntohl(reserved_r_status);

  if(GET_BITS_MIDDLE(reserved_r_status, 8, 1)/* r bit */)
    READ_DATA(&rep_relay_addr, recv_msg, length_read, sizeof(rep_relay_addr));

  if(GET_BITS_LOW(reserved_r_status, 7)) {
    warnx("Non-zero status field: %u in the reply message\n",
	   GET_BITS_LOW(reserved_r_status, 7));
    return 0;
  }

  /* If we are here it means success */

  while(length_read < n) {
    READ_DATA(&tl, recv_msg, length_read, sizeof(tl));
    tl.type      = ntohs(tl.type);
    tl.length    = ntohs(tl.length);

    switch(tl.type){
      
    case IP_ADDRESS_EXTENSION:
      printf("This should not occur. IP Address Extension in response to c bit in request\n");
      length_read += tl.length;
      break;
      
    default:
      /* Ignore the extension */
      printf("Extension in response to c bit in request. Ignoring\n");
      length_read += tl.length;
      
    }/* switch */
  }
  i = 0;

  while(*temp_liaddr) {
    if(!is_present(&(*temp_liaddr)->ip_addr, addr, length)) {
      t_liaddr     = *temp_liaddr;
      *temp_liaddr = (*temp_liaddr)->next;
      free(t_liaddr);
      i++;
      continue;
    }
    temp_liaddr = &(*temp_liaddr)->next;
  }
  return 1;
}

int send_and_process_request_any_addr(uint8_t A, uint8_t P, char *dns_name) {
  /* Gets an address with any scope or prefix len.
   * This function is like saying "Just get me an addr. I don't
   * care what the scope or prefix_length is". Note: scope will
   * always be global or sitelocal and if you are using dhcp inside
   * your site this function is useful. Also note that server just
   * doesn't pick up any arbitrary prefix_len. If you have a host
   * declaration that will be picked. Moreover only those prefixes
   * as specified in interface declaration (onlink clients) or subnet
   * declaration (offlink clients) are picked
   * Requests the first server in solicit list.
   */
  HEADER_CLIENTD_PACKET   header;
  LIADDR                **temp_liaddr = &liaddr_list_head, *t_liaddr;
  uint32_t                reserved_r_status;
  IP_ADDR_EXT            *iae_head = NULL, **temp_iae = &iae_head, *t_iae;
  struct in6_addr        *req_relay_addr, *req_server_addr, rep_relay_addr;
  TYPE_LEN                tl;
  TIME                    current_time;
  uint32_t                length_read;
  int                     i;

  bzero((char *) &rep_relay_addr, sizeof(rep_relay_addr));
  
  if(!ssi_head) {
    warnx("No server to send request to even after solicit. Returning...\n");
    return 0;
  }
  /* Build up a list of IP address extensions to be sent. These are not really
   * IP Address Extensions but a list of structures each of which can be expanded
   * into an IP Address Extension.
   */
  iae_head = (IP_ADDR_EXT *) malloc(sizeof(IP_ADDR_EXT));
  if(!iae_head)
    errx(1, "Insufficient memory.Exiting...\n");
  bzero((char *) iae_head, sizeof(*iae_head));

  iae_head->s.status     = 0;
  iae_head->s.C          = 0;
  iae_head->s.I          = 0;
  iae_head->s.L          = 0;
  iae_head->s.Q          = 1;
  iae_head->s.A          = A;
  iae_head->s.P          = P;
  iae_head->s.reserved   = 0;
  iae_head->s.scope      = 0;
  iae_head->s.prefix_len = 0;
  strcpy(iae_head->dns_name, dns_name);

  /* Imp
  iae_head->next = (IP_ADDR_EXT *) malloc(sizeof(IP_ADDR_EXT));
  if(!iae_head->next)
    errx(1, "Insufficient memory.Exiting...\n");
  bzero((char *) iae_head->next, sizeof(*iae_head->next));
  
  iae_head->next->s.status     = 0;
  iae_head->next->s.C          = 0;
  iae_head->next->s.I          = 0;
  iae_head->next->s.L          = 0;
  iae_head->next->s.Q          = 0;
  iae_head->next->s.A          = A;
  iae_head->next->s.P          = P;
  iae_head->next->s.reserved   = 0;
  iae_head->next->s.scope      = 0;
  iae_head->next->s.prefix_len = 0;
  
  strcpy(iae_head->next->dns_name, "Imp.com");
  */

  /* If you want to change the address of server to which you want to send a request
   * just change the following two fields
   */
  req_relay_addr  = &ssi_head->relay_addr;
  req_server_addr = &ssi_head->server_addr;

  printf("Sending request to client daemon\n");
  i = request(NULL, req_relay_addr, req_server_addr, 0 /* C */, 0 /* R */,
	      0/* T */, 0/* transaction_id */, iae_head, NULL, 0, NULL, 0);
  
  if(gettimeofday(&current_time, NULL) == -1)
    errx(1, "Error in gettimeofday. Exiting...\n");
  if(i == 0) {
    printf("Failure in getting an address. Returning...\n");
    return 0;
  }
  /* If we are here it means success */
  length_read = 0;
  while(*temp_liaddr)
    temp_liaddr = &((*temp_liaddr)->next);
  
  READ_DATA(&header, recv_msg, length_read, sizeof(header));
  header.pid        = ntohs(header.pid);
  header.request_no = ntohl(header.request_no);

  READ_DATA(&reserved_r_status, recv_msg, length_read, sizeof(reserved_r_status));
  reserved_r_status = ntohl(reserved_r_status);

  if(GET_BITS_MIDDLE(reserved_r_status, 8, 1)/* r bit */)
    READ_DATA(&rep_relay_addr, recv_msg, length_read, sizeof(rep_relay_addr));

  if(GET_BITS_LOW(reserved_r_status, 7)) {
    warnx("Non-zero status field: %u in the reply message\n",
	   GET_BITS_LOW(reserved_r_status, 7));
    return 0;
  }
  while(length_read < n) {
    READ_DATA(&tl, recv_msg, length_read, sizeof(tl));
    tl.type      = ntohs(tl.type);
    tl.length    = ntohs(tl.length);

    switch(tl.type){
      
    case IP_ADDRESS_EXTENSION:
      {
	struct{
	  uint32_t status     : 8;
	  uint32_t C          : 1;
	  uint32_t I          : 1;
	  uint32_t L          : 1;
	  uint32_t Q          : 1;
	  uint32_t A          : 1;
	  uint32_t P          : 1;
	  uint32_t reserved   : 8;
	  uint32_t scope      : 3;
	  uint32_t prefix_len : 7;
	} s;
	
	uint32_t         temp_ext;
	struct in6_addr  ip_addr;
	uint32_t         preferred_lifetime, valid_lifetime, old_length, name_size;
	char             rep_dns_name[MAX_DNS_NAME];
	
	name_size    = 0;
	old_length   = length_read;
	READ_DATA(&temp_ext, recv_msg, length_read, sizeof(temp_ext));
	
	temp_ext     = ntohl(temp_ext);
	s.status     = GET_BITS_HIGH(temp_ext, 8, 32);
	s.C          = GET_BITS_MIDDLE(temp_ext, 24, 1);
	s.I          = GET_BITS_MIDDLE(temp_ext, 23, 1);
	s.L          = GET_BITS_MIDDLE(temp_ext, 22, 1);
	s.Q          = GET_BITS_MIDDLE(temp_ext, 21, 1);
	s.A          = GET_BITS_MIDDLE(temp_ext, 20, 1);
	s.P          = GET_BITS_MIDDLE(temp_ext, 19, 1);
	s.reserved   = 0;
	s.scope      = GET_BITS_MIDDLE(temp_ext, 10, 3);
	s.prefix_len = GET_BITS_MIDDLE(temp_ext, 7, 7);
	
	if(s.status != 0) {
	  printf("Non-zero status %u in IP Address Extension. Discarding extension\n", s.status);
	  length_read = old_length + tl.length;
	  break;
	}

	if(s.C) {
	  READ_DATA(& ip_addr, recv_msg, length_read, sizeof(struct in6_addr));
	}
	else {
	  /* If no address is present extension is useless unless we ask
	   * for dns update of a non dhcp address which aren't doing. Skip
	   * the extension.
	   */
#ifdef DEBUG
	  printf("No IP address in IP_ADDRESS_EXTENSION in reply. "
		 "Ignoring the extension...\n");
#endif
	  length_read = old_length + tl.length;
	  break;
	}
	
	if(s.L){
	  READ_DATA(& preferred_lifetime, recv_msg, length_read, sizeof(preferred_lifetime));
	  READ_DATA(& valid_lifetime, recv_msg, length_read, sizeof(valid_lifetime));
	  preferred_lifetime = ntohl(preferred_lifetime);
	  valid_lifetime     = ntohl(valid_lifetime);
	}
	
	if((length_read - old_length) < tl.length){
	  /* This means DNS name is present. Read it off */
	  name_size           = tl.length - (length_read - old_length);
	  READ_DATA(rep_dns_name, recv_msg, length_read, name_size);
	  rep_dns_name[name_size] = '\0';
	  /* length_read = tl.length + old_length is also equally correct */
#ifdef DEBUG
	  printf("DNS name = %s in IP ADDRESS EXTENSION. This shouldn't occur\n", rep_dns_name);
#endif
	  
	}
	if(!s.L) {
	  /* If no lifetimes are present extension is useless (unless we ask
	   * for dns update of a non dhcp address which aren't doing) because client
	   * wouldn't be able to manage the lease.
	   */
#ifdef DEBUG
	  printf("No lifetimes in IP_ADDRESS_EXTENSION in reply. "
		 "Ignoring the extension...\n");
#endif
	  printf("Do we need to send a release\n");
	  {
	    struct in6_addr x_addr;
	    
	    if(inet_pton(AF_INET6, X_ADDR, &x_addr) <= 0)
	      errx("Error in converting address to binary\n");

	    send_and_process_release_by_extension(recv_msg + old_length - sizeof(tl),
						  tl.length, 0, req_server_addr,
						  &x_addr);
	  }
	  length_read = old_length + tl.length;
	  break;
	}
	
	/* Address and lifetimes are present. We can use the address */
	t_liaddr                     = (LIADDR *) malloc(sizeof(LIADDR));
	if(!t_liaddr)
	  errx(1, "Insufficient memory. Exiting...\n");
	bzero((char *) t_liaddr, sizeof(*t_liaddr));

	t_liaddr->ip_addr            = ip_addr;
	strcpy(t_liaddr->dns_name, dns_name);
	t_liaddr->relay_addr         = rep_relay_addr;
	t_liaddr->server_addr        = *req_server_addr;
	t_liaddr->when_leased        = current_time;
	t_liaddr->preferred_lifetime = preferred_lifetime;
	t_liaddr->valid_lifetime     = valid_lifetime;
	t_liaddr->A                  = s.A;
	t_liaddr->P                  = s.P;

	*temp_liaddr                 = t_liaddr;
	temp_liaddr                  = &((*temp_liaddr)->next);
      }/* case IP_ADDRESS EXT */
      break;
      
    default:
      /* Ignore the extension */
      length_read += tl.length;
      
    }/* switch */
  }
  return 1;
}

int send_and_process_request_any_addr_with_lifetime(uint32_t preferred_lifetime,
						    uint32_t valid_lifetime,
						    uint8_t A, uint8_t P, char *dns_name) {
  
  HEADER_CLIENTD_PACKET   header;
  LIADDR                **temp_liaddr = &liaddr_list_head, *t_liaddr;
  uint32_t                reserved_r_status;
  IP_ADDR_EXT            *iae_head = NULL, **temp_iae = &iae_head, *t_iae;
  struct in6_addr        *req_relay_addr, *req_server_addr, rep_relay_addr;
  TYPE_LEN                tl;
  TIME                    current_time;
  uint32_t                length_read;
  int                     i;

  bzero((char *) &rep_relay_addr, sizeof(rep_relay_addr));
  
  if(!ssi_head) {
    warnx("No server to send request to even after solicit. Returning...\n");
    return 0;
  }
  /* Build up a list of IP address extensions to be sent. These are not really
   * IP Address Extensions but a list of structures each of which can be expanded
   * into an IP Address Extension.
   */
  iae_head = (IP_ADDR_EXT *) malloc(sizeof(IP_ADDR_EXT));
  if(!iae_head)
    errx(1, "Insufficient memory.Exiting...\n");
  bzero((char *) iae_head, sizeof(*iae_head));

  iae_head->s.status     = 0;
  iae_head->s.C          = 0;
  iae_head->s.I          = 0;
  iae_head->s.L          = 1;
  iae_head->s.Q          = 1;
  iae_head->s.A          = A;
  iae_head->s.P          = P;
  iae_head->s.reserved   = 0;
  iae_head->s.scope      = 0;
  iae_head->s.prefix_len = 0;
  iae_head->preferred_lifetime = preferred_lifetime;
  iae_head->valid_lifetime     = valid_lifetime;

  strcpy(iae_head->dns_name, dns_name);

  req_relay_addr  = &ssi_head->relay_addr;
  req_server_addr = &ssi_head->server_addr;

  printf("Sending request to client daemon\n");
  i = request(NULL, req_relay_addr, req_server_addr, 0 /* C */, 0 /* R */,
	      0/* T */, 0/* transaction_id */, iae_head, NULL, 0, NULL, 0);
  
  if(gettimeofday(&current_time, NULL) == -1)
    errx(1, "Error in gettimeofday. Exiting...\n");
  if(i == 0) {
    printf("Failure in getting an address. Returning...\n");
    return 0;
  }
  /* If we are here it means success */
  length_read = 0;
  while(*temp_liaddr)
    temp_liaddr = &((*temp_liaddr)->next);
  
  READ_DATA(&header, recv_msg, length_read, sizeof(header));
  header.pid        = ntohs(header.pid);
  header.request_no = ntohl(header.request_no);

  READ_DATA(&reserved_r_status, recv_msg, length_read, sizeof(reserved_r_status));
  reserved_r_status = ntohl(reserved_r_status);

  if(GET_BITS_MIDDLE(reserved_r_status, 8, 1)/* r bit */)
    READ_DATA(&rep_relay_addr, recv_msg, length_read, sizeof(rep_relay_addr));

  if(GET_BITS_LOW(reserved_r_status, 7)) {
    warnx("Non-zero status field: %u in the reply message\n",
	   GET_BITS_LOW(reserved_r_status, 7));
    return 0;
  }
  while(length_read < n) {
    READ_DATA(&tl, recv_msg, length_read, sizeof(tl));
    tl.type      = ntohs(tl.type);
    tl.length    = ntohs(tl.length);

    switch(tl.type){
      
    case IP_ADDRESS_EXTENSION:
      {
	struct{
	  uint32_t status     : 8;
	  uint32_t C          : 1;
	  uint32_t I          : 1;
	  uint32_t L          : 1;
	  uint32_t Q          : 1;
	  uint32_t A          : 1;
	  uint32_t P          : 1;
	  uint32_t reserved   : 8;
	  uint32_t scope      : 3;
	  uint32_t prefix_len : 7;
	} s;
	
	uint32_t         temp_ext;
	struct in6_addr  ip_addr;
	uint32_t         preferred_lifetime, valid_lifetime, old_length, name_size;
	char             rep_dns_name[MAX_DNS_NAME];
	
	name_size    = 0;
	old_length   = length_read;
	READ_DATA(&temp_ext, recv_msg, length_read, sizeof(temp_ext));
	
	temp_ext     = ntohl(temp_ext);
	s.status     = GET_BITS_HIGH(temp_ext, 8, 32);
	s.C          = GET_BITS_MIDDLE(temp_ext, 24, 1);
	s.I          = GET_BITS_MIDDLE(temp_ext, 23, 1);
	s.L          = GET_BITS_MIDDLE(temp_ext, 22, 1);
	s.Q          = GET_BITS_MIDDLE(temp_ext, 21, 1);
	s.A          = GET_BITS_MIDDLE(temp_ext, 20, 1);
	s.P          = GET_BITS_MIDDLE(temp_ext, 19, 1);
	s.reserved   = 0;
	s.scope      = GET_BITS_MIDDLE(temp_ext, 10, 3);
	s.prefix_len = GET_BITS_MIDDLE(temp_ext, 7, 7);
	
	if(s.status != 0) {
	  printf("Non-zero status %u in IP Address Extension. Discarding extension\n", s.status);
	  length_read = old_length + tl.length;
	  break;
	}

	if(s.C) {
	  READ_DATA(& ip_addr, recv_msg, length_read, sizeof(struct in6_addr));
	}
	else {
	  /* If no address is present extension is useless unless we ask
	   * for dns update of a non dhcp address which aren't doing. Skip
	   * the extension.
	   */
#ifdef DEBUG
	  printf("No IP address in IP_ADDRESS_EXTENSION in reply. "
		 "Ignoring the extension...\n");
#endif
	  length_read = old_length + tl.length;
	  break;
	}
	
	if(s.L){
	  READ_DATA(& preferred_lifetime, recv_msg, length_read, sizeof(preferred_lifetime));
	  READ_DATA(& valid_lifetime, recv_msg, length_read, sizeof(valid_lifetime));
	  preferred_lifetime = ntohl(preferred_lifetime);
	  valid_lifetime     = ntohl(valid_lifetime);
	}
	
	if((length_read - old_length) < tl.length){
	  /* This means DNS name is present. Read it off */
	  name_size           = tl.length - (length_read - old_length);
	  READ_DATA(rep_dns_name, recv_msg, length_read, name_size);
	  rep_dns_name[name_size] = '\0';
	  /* length_read = tl.length + old_length is also equally correct */
#ifdef DEBUG
	  printf("DNS name = %s in IP ADDRESS EXTENSION. This shouldn't occur\n", rep_dns_name);
#endif
	  
	}
	if(!s.L) {
	  /* If no lifetimes are present extension is useless (unless we ask
	   * for dns update of a non dhcp address which aren't doing) because client
	   * wouldn't be able to manage the lease.
	   */
#ifdef DEBUG
	  printf("No lifetimes in IP_ADDRESS_EXTENSION in reply. "
		 "Ignoring the extension...\n");
#endif
	  printf("Do we need to send a release\n");
	  {
	    struct in6_addr x_addr;
	    
	    if(inet_pton(AF_INET6, X_ADDR, &x_addr) <= 0)
	      errx("Error in converting address to binary\n");

	    send_and_process_release_by_extension(recv_msg + old_length - sizeof(tl),
						  tl.length, 0, req_server_addr,
						  &x_addr);
	  }
	  length_read = old_length + tl.length;
	  break;
	}
	
	/* Address and lifetimes are present. We can use the address */
	t_liaddr                     = (LIADDR *) malloc(sizeof(LIADDR));
	if(!t_liaddr)
	  errx(1, "Insufficient memory. Exiting...\n");
	bzero((char *) t_liaddr, sizeof(*t_liaddr));
	t_liaddr->ip_addr            = ip_addr;
	strcpy(t_liaddr->dns_name, dns_name);
	t_liaddr->relay_addr         = rep_relay_addr;
	t_liaddr->server_addr        = *req_server_addr;
	t_liaddr->when_leased        = current_time;
	t_liaddr->preferred_lifetime = preferred_lifetime;
	t_liaddr->valid_lifetime     = valid_lifetime;
	t_liaddr->A                  = s.A;
	t_liaddr->P                  = s.P;

	*temp_liaddr                 = t_liaddr;
	temp_liaddr                  = &((*temp_liaddr)->next);
      }/* case IP_ADDRESS EXT */
      break;
      
    default:
      /* Ignore the extension */
      length_read += tl.length;
      
    }/* switch */
  }
  return 1;
}

int send_and_process_request_specific_addr(struct in6_addr *ip_addr, uint8_t A, uint8_t P,
					   char *dns_name, int pre_len) {
  /* Gets a specific addr with any scope or prefix len
   * Requests the first server in solicit list.
   */
  HEADER_CLIENTD_PACKET   header;
  LIADDR                **temp_liaddr = &liaddr_list_head, *t_liaddr;
  uint32_t                reserved_r_status;
  IP_ADDR_EXT            *iae_head = NULL, **temp_iae = &iae_head, *t_iae;
  struct in6_addr        *req_relay_addr, *req_server_addr, rep_relay_addr;
  TYPE_LEN                tl;
  TIME                    current_time;
  uint32_t                length_read;
  int                     i;

  bzero((char *) &rep_relay_addr, sizeof(rep_relay_addr));
  
  if(!ssi_head) {
    warnx("No server to send request to even after solicit. Returning...\n");
    return 0;
  }
  /* Build up a list of IP address extensions to be sent. These are not really
   * IP Address Extensions but a list of structures each of which can be expanded
   * into an IP Address Extension.
   */
  iae_head = (IP_ADDR_EXT *) malloc(sizeof(IP_ADDR_EXT));
  if(!iae_head)
    errx(1, "Insufficient memory.Exiting...\n");
  bzero((char *) iae_head, sizeof(*iae_head));

  iae_head->s.status     = 0;
  iae_head->s.C          = 1;
  iae_head->s.I          = 0;
  iae_head->s.L          = 0;
  iae_head->s.Q          = 1;
  iae_head->s.A          = A;
  iae_head->s.P          = P;
  iae_head->s.reserved   = 0;
  iae_head->s.scope      = 0;
  iae_head->s.prefix_len = pre_len;
  iae_head->ip_addr      = *ip_addr;
  strcpy(iae_head->dns_name, dns_name);

  req_relay_addr  = &ssi_head->relay_addr;
  req_server_addr = &ssi_head->server_addr;

  printf("Sending request to client daemon\n");
  i = request(NULL, req_relay_addr, req_server_addr, 0 /* C */, 0 /* R */,
	      0/* T */, 0/* transaction_id */, iae_head, NULL, 0, NULL, 0);
  
  if(gettimeofday(&current_time, NULL) == -1)
    errx(1, "Error in gettimeofday. Exiting...\n");
  if(i == 0) {
    printf("Failure in getting an address. Returning...\n");
    return 0;
  }
  /* If we are here it means success */
  length_read = 0;
  while(*temp_liaddr)
    temp_liaddr = &((*temp_liaddr)->next);
  
  READ_DATA(&header, recv_msg, length_read, sizeof(header));
  header.pid        = ntohs(header.pid);
  header.request_no = ntohl(header.request_no);

  READ_DATA(&reserved_r_status, recv_msg, length_read, sizeof(reserved_r_status));
  reserved_r_status = ntohl(reserved_r_status);

  if(GET_BITS_MIDDLE(reserved_r_status, 8, 1)/* r bit */)
    READ_DATA(&rep_relay_addr, recv_msg, length_read, sizeof(rep_relay_addr));

  if(GET_BITS_LOW(reserved_r_status, 7)) {
    warnx("Non-zero status field: %u in the reply message\n",
	   GET_BITS_LOW(reserved_r_status, 7));
    return 0;
  }
  while(length_read < n) {
    READ_DATA(&tl, recv_msg, length_read, sizeof(tl));
    tl.type      = ntohs(tl.type);
    tl.length    = ntohs(tl.length);

    switch(tl.type){
      
    case IP_ADDRESS_EXTENSION:
      {
	struct{
	  uint32_t status     : 8;
	  uint32_t C          : 1;
	  uint32_t I          : 1;
	  uint32_t L          : 1;
	  uint32_t Q          : 1;
	  uint32_t A          : 1;
	  uint32_t P          : 1;
	  uint32_t reserved   : 8;
	  uint32_t scope      : 3;
	  uint32_t prefix_len : 7;
	} s;
	
	uint32_t         temp_ext;
	struct in6_addr  ip_addr;
	uint32_t         preferred_lifetime, valid_lifetime, old_length, name_size;
	char             rep_dns_name[MAX_DNS_NAME];
	
	name_size    = 0;
	old_length   = length_read;
	READ_DATA(&temp_ext, recv_msg, length_read, sizeof(temp_ext));
	
	temp_ext     = ntohl(temp_ext);
	s.status     = GET_BITS_HIGH(temp_ext, 8, 32);
	s.C          = GET_BITS_MIDDLE(temp_ext, 24, 1);
	s.I          = GET_BITS_MIDDLE(temp_ext, 23, 1);
	s.L          = GET_BITS_MIDDLE(temp_ext, 22, 1);
	s.Q          = GET_BITS_MIDDLE(temp_ext, 21, 1);
	s.A          = GET_BITS_MIDDLE(temp_ext, 20, 1);
	s.P          = GET_BITS_MIDDLE(temp_ext, 19, 1);
	s.reserved   = 0;
	s.scope      = GET_BITS_MIDDLE(temp_ext, 10, 3);
	s.prefix_len = GET_BITS_MIDDLE(temp_ext, 7, 7);
	
	if(s.status != 0) {
	  printf("Non-zero status %u in IP Address Extension. Discarding extension\n", s.status);
	  length_read = old_length + tl.length;
	  break;
	}
	if(s.C) {
	  READ_DATA(& ip_addr, recv_msg, length_read, sizeof(struct in6_addr));
	}
	else {
	  /* If no address is present extension is useless unless we ask
	   * for dns update of a non dhcp address which aren't doing. Skip
	   * the extension.
	   */
#ifdef DEBUG
	  printf("No IP address in IP_ADDRESS_EXTENSION in reply. "
		 "Ignoring the extension...\n");
#endif
	  length_read = old_length + tl.length;
	  break;
	}
	
	if(s.L){
	  READ_DATA(& preferred_lifetime, recv_msg, length_read, sizeof(preferred_lifetime));
	  READ_DATA(& valid_lifetime, recv_msg, length_read, sizeof(valid_lifetime));
	  preferred_lifetime = ntohl(preferred_lifetime);
	  valid_lifetime     = ntohl(valid_lifetime);
	}
	
	if((length_read - old_length) < tl.length){
	  /* This means DNS name is present. Read it off */
	  name_size           = tl.length - (length_read - old_length);
	  READ_DATA(rep_dns_name, recv_msg, length_read, name_size);
	  rep_dns_name[name_size] = '\0';
	  /* length_read = tl.length + old_length is also equally correct */
#ifdef DEBUG
	  printf("DNS name = %s in IP ADDRESS EXTENSION. This shouldn't occur\n", rep_dns_name);
#endif
	  
	}
	if(!s.L) {
	  /* If no lifetimes are present extension is useless (unless we ask
	   * for dns update of a non dhcp address which aren't doing) because client
	   * wouldn't be able to manage the lease.
	   */
#ifdef DEBUG
	  printf("No lifetimes in IP_ADDRESS_EXTENSION in reply. "
		 "Ignoring the extension...\n");
#endif
	  printf("Do we need to send a release\n");
	  {
	    struct in6_addr x_addr;
	    
	    if(inet_pton(AF_INET6, X_ADDR, &x_addr) <= 0)
	      errx("Error in converting address to binary\n");

	    send_and_process_release_by_extension(recv_msg + old_length - sizeof(tl),
						  tl.length, 0, req_server_addr,
						  &x_addr);
	  }
	  length_read = old_length + tl.length;
	  break;
	}
	
	/* Address and lifetimes are present. We can use the address */
	t_liaddr                     = (LIADDR *) malloc(sizeof(LIADDR));
	if(!t_liaddr)
	  errx(1, "Insufficient memory. Exiting...\n");
	bzero((char *) t_liaddr, sizeof(*t_liaddr));
	t_liaddr->ip_addr            = ip_addr;
	strcpy(t_liaddr->dns_name, dns_name);
	t_liaddr->relay_addr         = rep_relay_addr;
	t_liaddr->server_addr        = *req_server_addr;
	t_liaddr->when_leased        = current_time;
	t_liaddr->preferred_lifetime = preferred_lifetime;
	t_liaddr->valid_lifetime     = valid_lifetime;
	t_liaddr->A                  = s.A;
	t_liaddr->P                  = s.P;

	*temp_liaddr                 = t_liaddr;
	temp_liaddr                  = &((*temp_liaddr)->next);
      }/* case IP_ADDRESS EXT */
      break;
      
    default:
      /* Ignore the extension */
      length_read += tl.length;
      
    }/* switch */
  }
  return 1;
}


int send_and_process_request_specific_addr_with_lifetime(uint32_t preferred_lifetime,
							 uint32_t valid_lifetime,
							 struct in6_addr *ip_addr,
							 uint8_t A, uint8_t P,
							 char *dns_name) {
  /* Gets a specific addr with any scope or prefix len
   * Requests the first server in solicit list.
   */
  HEADER_CLIENTD_PACKET   header;
  LIADDR                **temp_liaddr = &liaddr_list_head, *t_liaddr;
  uint32_t                reserved_r_status;
  IP_ADDR_EXT            *iae_head = NULL, **temp_iae = &iae_head, *t_iae;
  struct in6_addr        *req_relay_addr, *req_server_addr, rep_relay_addr;
  TYPE_LEN                tl;
  TIME                    current_time;
  uint32_t                length_read;
  int                     i;

  bzero((char *) &rep_relay_addr, sizeof(rep_relay_addr));
  
  if(!ssi_head) {
    warnx("No server to send request to even after solicit. Returning...\n");
    return 0;
  }
  /* Build up a list of IP address extensions to be sent. These are not really
   * IP Address Extensions but a list of structures each of which can be expanded
   * into an IP Address Extension.
   */
  iae_head = (IP_ADDR_EXT *) malloc(sizeof(IP_ADDR_EXT));
  if(!iae_head)
    errx(1, "Insufficient memory.Exiting...\n");
  bzero((char *) iae_head, sizeof(*iae_head));

  iae_head->s.status     = 0;
  iae_head->s.C          = 1;
  iae_head->s.I          = 0;
  iae_head->s.L          = 1;
  iae_head->s.Q          = 1;
  iae_head->s.A          = A;
  iae_head->s.P          = P;
  iae_head->s.reserved   = 0;
  iae_head->s.scope      = 0;
  iae_head->s.prefix_len = 0;
  iae_head->ip_addr      = *ip_addr;
  iae_head->preferred_lifetime = preferred_lifetime;
  iae_head->valid_lifetime     = valid_lifetime;
  strcpy(iae_head->dns_name, dns_name);

  req_relay_addr  = &ssi_head->relay_addr;
  req_server_addr = &ssi_head->server_addr;

  printf("Sending request to client daemon\n");
  i = request(NULL, req_relay_addr, req_server_addr, 0 /* C */, 0 /* R */,
	      0/* T */, 0/* transaction_id */, iae_head, NULL, 0, NULL, 0);
  
  if(gettimeofday(&current_time, NULL) == -1)
    errx(1, "Error in gettimeofday. Exiting...\n");
  if(i == 0) {
    printf("Failure in getting an address. Returning...\n");
    return 0;
  }
  /* If we are here it means success */
  length_read = 0;
  while(*temp_liaddr)
    temp_liaddr = &((*temp_liaddr)->next);
  
  READ_DATA(&header, recv_msg, length_read, sizeof(header));
  header.pid        = ntohs(header.pid);
  header.request_no = ntohl(header.request_no);

  READ_DATA(&reserved_r_status, recv_msg, length_read, sizeof(reserved_r_status));
  reserved_r_status = ntohl(reserved_r_status);

  if(GET_BITS_MIDDLE(reserved_r_status, 8, 1)/* r bit */)
    READ_DATA(&rep_relay_addr, recv_msg, length_read, sizeof(rep_relay_addr));

  if(GET_BITS_LOW(reserved_r_status, 7)) {
    warnx("Non-zero status field: %u in the reply message\n",
	   GET_BITS_LOW(reserved_r_status, 7));
    return 0;
  }
  while(length_read < n) {
    READ_DATA(&tl, recv_msg, length_read, sizeof(tl));
    tl.type      = ntohs(tl.type);
    tl.length    = ntohs(tl.length);

    switch(tl.type){
      
    case IP_ADDRESS_EXTENSION:
      {
	struct{
	  uint32_t status     : 8;
	  uint32_t C          : 1;
	  uint32_t I          : 1;
	  uint32_t L          : 1;
	  uint32_t Q          : 1;
	  uint32_t A          : 1;
	  uint32_t P          : 1;
	  uint32_t reserved   : 8;
	  uint32_t scope      : 3;
	  uint32_t prefix_len : 7;
	} s;
	
	uint32_t         temp_ext;
	struct in6_addr  ip_addr;
	uint32_t         preferred_lifetime, valid_lifetime, old_length, name_size;
	char             rep_dns_name[MAX_DNS_NAME];
	
	name_size    = 0;
	old_length   = length_read;
	READ_DATA(&temp_ext, recv_msg, length_read, sizeof(temp_ext));
	
	temp_ext     = ntohl(temp_ext);
	s.status     = GET_BITS_HIGH(temp_ext, 8, 32);
	s.C          = GET_BITS_MIDDLE(temp_ext, 24, 1);
	s.I          = GET_BITS_MIDDLE(temp_ext, 23, 1);
	s.L          = GET_BITS_MIDDLE(temp_ext, 22, 1);
	s.Q          = GET_BITS_MIDDLE(temp_ext, 21, 1);
	s.A          = GET_BITS_MIDDLE(temp_ext, 20, 1);
	s.P          = GET_BITS_MIDDLE(temp_ext, 19, 1);
	s.reserved   = 0;
	s.scope      = GET_BITS_MIDDLE(temp_ext, 10, 3);
	s.prefix_len = GET_BITS_MIDDLE(temp_ext, 7, 7);
	
	if(s.status != 0) {
	  printf("Non-zero status %u in IP Address Extension. Discarding extension\n", s.status);
	  length_read = old_length + tl.length;
	  break;
	}

	if(s.C) {
	  READ_DATA(& ip_addr, recv_msg, length_read, sizeof(struct in6_addr));
	}
	else {
	  /* If no address is present extension is useless unless we ask
	   * for dns update of a non dhcp address which aren't doing. Skip
	   * the extension.
	   */
#ifdef DEBUG
	  printf("No IP address in IP_ADDRESS_EXTENSION in reply. "
		 "Ignoring the extension...\n");
#endif
	  length_read = old_length + tl.length;
	  break;
	}
	
	if(s.L){
	  READ_DATA(& preferred_lifetime, recv_msg, length_read, sizeof(preferred_lifetime));
	  READ_DATA(& valid_lifetime, recv_msg, length_read, sizeof(valid_lifetime));
	  preferred_lifetime = ntohl(preferred_lifetime);
	  valid_lifetime     = ntohl(valid_lifetime);
	}
	
	if((length_read - old_length) < tl.length){
	  /* This means DNS name is present. Read it off */
	  name_size           = tl.length - (length_read - old_length);
	  READ_DATA(rep_dns_name, recv_msg, length_read, name_size);
	  rep_dns_name[name_size] = '\0';
	  /* length_read = tl.length + old_length is also equally correct */
#ifdef DEBUG
	  printf("DNS name = %s in IP ADDRESS EXTENSION. This shouldn't occur\n", rep_dns_name);
#endif
	  
	}
	if(!s.L) {
	  /* If no lifetimes are present extension is useless (unless we ask
	   * for dns update of a non dhcp address which aren't doing) because client
	   * wouldn't be able to manage the lease.
	   */
#ifdef DEBUG
	  printf("No lifetimes in IP_ADDRESS_EXTENSION in reply. "
		 "Ignoring the extension...\n");
#endif
	  printf("Do we need to send a release\n");
	  {
	    struct in6_addr x_addr;
	    
	    if(inet_pton(AF_INET6, X_ADDR, &x_addr) <= 0)
	      errx("Error in converting address to binary\n");

	    send_and_process_release_by_extension(recv_msg + old_length - sizeof(tl),
						  tl.length, 0, req_server_addr,
						  &x_addr);
	  }
	  length_read = old_length + tl.length;
	  break;
	}
	
	/* Address and lifetimes are present. We can use the address */
	t_liaddr                     = (LIADDR *) malloc(sizeof(LIADDR));
	if(!t_liaddr)
	  errx(1, "Insufficient memory. Exiting...\n");
	bzero((char *) t_liaddr, sizeof(*t_liaddr));
	t_liaddr->ip_addr            = ip_addr;
	strcpy(t_liaddr->dns_name, dns_name);
	t_liaddr->relay_addr         = rep_relay_addr;
	t_liaddr->server_addr        = *req_server_addr;
	t_liaddr->when_leased        = current_time;
	t_liaddr->preferred_lifetime = preferred_lifetime;
	t_liaddr->valid_lifetime     = valid_lifetime;
	t_liaddr->A                  = s.A;
	t_liaddr->P                  = s.P;

	*temp_liaddr                 = t_liaddr;
	temp_liaddr                  = &((*temp_liaddr)->next);
      }/* case IP_ADDRESS EXT */
      break;
      
    default:
      /* Ignore the extension */
      length_read += tl.length;
      
    }/* switch */
  }
  return 1;
}


int send_and_process_release(uint8_t A, uint8_t P, uint8_t Q,
			     struct in6_addr *ip_addr, uint8_t r,
			     struct in6_addr *server_addr,
			     struct in6_addr *x_addr) {
  HEADER_CLIENTD_PACKET   header;
  IP_ADDR_EXT            *iae_head = NULL, **temp_iae = &iae_head, *t_iae;
  int                     i;
  LIADDR                **temp_liaddr = &liaddr_list_head, *t_liaddr;
  uint32_t                length_read;
  char                    name[64];


  /* Check if ip_addr is actually present in liaddr_list */
  for(; *temp_liaddr; temp_liaddr = &(*temp_liaddr)->next) {
    if(IN6_ARE_ADDR_EQUAL(&(*temp_liaddr)->ip_addr, ip_addr))
      break;
  }
  if(!(*temp_liaddr)) {
    warnx("Could not find address %s to be released. Returning...\n",
	  inet_ntop(AF_INET6, ip_addr, name, 64));
    return 0;
  }

  if(!server_addr || IN6_IS_ADDR_UNSPECIFIED(server_addr)) {
    warnx("NULL or unspecified server addr. Returning...\n");
    return 0;
  }

  if(!x_addr || IN6_IS_ADDR_UNSPECIFIED(x_addr)) {
    warnx("NULL or unspecified server addr. Returning...\n");
    return 0;
  }

  if(!r && IN6_IS_ADDR_LINKLOCAL(x_addr)) {
    /* If r bit is not set x_addr is of non-local-scope */
    warnx("x_addr is of non-local-scope. Returning...\n");
    return 0;
  }

  /* Build up a list of IP address extensions to be sent. These are not really
   * IP Address Extensions but a list of structures each of which can be expanded
   * into an IP Address Extension.
   */

  iae_head = (IP_ADDR_EXT *) malloc(sizeof(IP_ADDR_EXT));
  if(!iae_head)
    errx(1, "Insufficient memory.Exiting...\n");
  bzero((char *) iae_head, sizeof(*iae_head));

  iae_head->s.status     = 0;
  iae_head->s.C          = 1;
  iae_head->s.I          = 0;
  iae_head->s.L          = 0;
  iae_head->s.Q          = Q;
  iae_head->s.A          = A;
  iae_head->s.P          = P;
  iae_head->s.reserved   = 0;
  iae_head->s.scope      = 0;
  iae_head->s.prefix_len = 0;
  iae_head->dns_name[0]  = 0;
  iae_head->ip_addr      = *ip_addr;

  i = release(r, server_addr, x_addr, iae_head);

  if(i == 0) {
    printf("Failure in releasing %s. Returning...\n", inet_ntop(AF_INET6, ip_addr, name, 64));
    return 0;
  }
  printf("Successfully released %s.\n", inet_ntop(AF_INET6, ip_addr, name, 64));
  /* If we are here it means success */
  length_read = 0;
  /* Check the reply to see if actually the address was released */

  /* Nevertheless delete the address from the list */
  if(A)
    (*temp_liaddr)->A = 0;
  if(P)
    (*temp_liaddr)->P = 0;
  
  if(!A && !P) {
    t_liaddr = *temp_liaddr;
    *temp_liaddr = (*temp_liaddr)->next;
    printf("Deleting %s from leased addr list.\n",
	   inet_ntop(AF_INET6, ip_addr, name, 64));
    free(t_liaddr);
  }
  return 1;
}

int send_and_process_dns_update(uint8_t A, uint8_t P, struct in6_addr *req_ip_addr,
				char *dns_name) {
  HEADER_CLIENTD_PACKET   header;
  uint32_t                reserved_r_status;
  IP_ADDR_EXT            *iae_head = NULL, **temp_iae = &iae_head, *t_iae;
  struct in6_addr        *req_relay_addr, *req_server_addr, rep_relay_addr;
  TYPE_LEN                tl;
  uint32_t                length_read;
  int                     i;
  uint8_t                 flag = 0;
  char                    name[64];

  bzero((char *) &rep_relay_addr, sizeof(rep_relay_addr));
  
  if(!ssi_head) {
    warnx("No server to send request to even after solicit. Returning...\n");
    return 0;
  }
  /* Build up a list of IP address extensions to be sent. These are not really
   * IP Address Extensions but a list of structures each of which can be expanded
   * into an IP Address Extension.
   */
  iae_head = (IP_ADDR_EXT *) malloc(sizeof(IP_ADDR_EXT));
  if(!iae_head)
    errx(1, "Insufficient memory.Exiting...\n");
  bzero((char *) iae_head, sizeof(*iae_head));

  iae_head->s.status     = 0;
  iae_head->s.C          = 1;
  iae_head->s.I          = 1;
  iae_head->s.L          = 0;
  iae_head->s.Q          = 1;// Imp: You would like to change this 
  iae_head->s.A          = A;
  iae_head->s.P          = P;
  iae_head->s.reserved   = 0;
  iae_head->s.scope      = 0;
  iae_head->s.prefix_len = 0;
  iae_head->ip_addr      = *req_ip_addr;
  strcpy(iae_head->dns_name, dns_name);

  req_relay_addr  = &ssi_head->relay_addr;
  req_server_addr = &ssi_head->server_addr;

  printf("Sending request to client daemon\n");
  i = request(NULL, req_relay_addr, req_server_addr, 0 /* C */, 0 /* R */,
	      0/* T */, 0/* transaction_id */, iae_head, NULL, 0, NULL, 0);
  if(i == 0) {
    printf("Failure in updating dns for %s, dns_name %s. Returning...\n",
	   inet_ntop(AF_INET6, req_ip_addr, name, 64),
	   dns_name);
    return 0;
  }

  length_read = 0;
  
  READ_DATA(&header, recv_msg, length_read, sizeof(header));
  header.pid        = ntohs(header.pid);
  header.request_no = ntohl(header.request_no);

  READ_DATA(&reserved_r_status, recv_msg, length_read, sizeof(reserved_r_status));
  reserved_r_status = ntohl(reserved_r_status);

  if(GET_BITS_MIDDLE(reserved_r_status, 8, 1)/* r bit */)
    READ_DATA(&rep_relay_addr, recv_msg, length_read, sizeof(rep_relay_addr));

  if(GET_BITS_LOW(reserved_r_status, 7)) {
    warnx("Non-zero status field: %u in the reply message\n",
	   GET_BITS_LOW(reserved_r_status, 7));
    return 0;
  }
  while(length_read < n) {
    READ_DATA(&tl, recv_msg, length_read, sizeof(tl));
    tl.type      = ntohs(tl.type);
    tl.length    = ntohs(tl.length);

    switch(tl.type){
      
    case IP_ADDRESS_EXTENSION:
      {
	struct{
	  uint32_t status     : 8;
	  uint32_t C          : 1;
	  uint32_t I          : 1;
	  uint32_t L          : 1;
	  uint32_t Q          : 1;
	  uint32_t A          : 1;
	  uint32_t P          : 1;
	  uint32_t reserved   : 8;
	  uint32_t scope      : 3;
	  uint32_t prefix_len : 7;
	} s;
	
	uint32_t         temp_ext;
	struct in6_addr  ip_addr;
	uint32_t         preferred_lifetime, valid_lifetime, old_length, name_size;
	char             rep_dns_name[MAX_DNS_NAME];
	
	name_size    = 0;
	old_length   = length_read;
	READ_DATA(&temp_ext, recv_msg, length_read, sizeof(temp_ext));
	
	temp_ext     = ntohl(temp_ext);
	s.status     = GET_BITS_HIGH(temp_ext, 8, 32);
	s.C          = GET_BITS_MIDDLE(temp_ext, 24, 1);
	s.I          = GET_BITS_MIDDLE(temp_ext, 23, 1);
	s.L          = GET_BITS_MIDDLE(temp_ext, 22, 1);
	s.Q          = GET_BITS_MIDDLE(temp_ext, 21, 1);
	s.A          = GET_BITS_MIDDLE(temp_ext, 20, 1);
	s.P          = GET_BITS_MIDDLE(temp_ext, 19, 1);
	s.reserved   = 0;
	s.scope      = GET_BITS_MIDDLE(temp_ext, 10, 3);
	s.prefix_len = GET_BITS_MIDDLE(temp_ext, 7, 7);
	
	if(s.status != 0) {
	  printf("Non-zero status %u in IP Address Extension. Discarding extension\n", s.status);
	  length_read = old_length + tl.length;
	  break;
	}
	if(s.C) {
	  READ_DATA(& ip_addr, recv_msg, length_read, sizeof(struct in6_addr));
	}
	else {
	  /* If no address is present extension is useless unless we ask
	   * for dns update of a non dhcp address which aren't doing. Skip
	   * the extension.
	   */
#ifdef DEBUG
	  printf("No IP address in IP_ADDRESS_EXTENSION in reply. "
		 "Ignoring the extension...\n");
#endif
	  length_read = old_length + tl.length;
	  break;
	}
	
	if(s.L){
	  READ_DATA(& preferred_lifetime, recv_msg, length_read, sizeof(preferred_lifetime));
	  READ_DATA(& valid_lifetime, recv_msg, length_read, sizeof(valid_lifetime));
	  preferred_lifetime = ntohl(preferred_lifetime);
	  valid_lifetime     = ntohl(valid_lifetime);
	}
	
	if((length_read - old_length) < tl.length){
	  /* This means DNS name is present. Read it off */
	  name_size           = tl.length - (length_read - old_length);
	  READ_DATA(rep_dns_name, recv_msg, length_read, name_size);
	  rep_dns_name[name_size] = '\0';
	  /* length_read = tl.length + old_length is also equally correct */
#ifdef DEBUG
	  printf("DNS name = %s in IP ADDRESS EXTENSION. This shouldn't occur\n", rep_dns_name);
#endif
	  
	}
	if(IN6_ARE_ADDR_EQUAL(&ip_addr, req_ip_addr)) {
	  printf("Update values for A and P %u %u\n", s.A, s.P);
	  flag = 1;
	}
      }/* case IP_ADDRESS EXT */
      break;
      
    default:
      /* Ignore the extension */
      length_read += tl.length;
      
    }/* switch */
  }

  if(!flag) {
    printf("Error in updating dns for non-dhcp addr %s, dns_name %s\n",
	   inet_ntop(AF_INET6, req_ip_addr, name, 64),
	   dns_name);
    return 0;
  }

  printf("Successfully updated dns for %s, dns_name %s. Returning...\n",
	 inet_ntop(AF_INET6, req_ip_addr, name, 64),
	 dns_name);
  return 1;
  
}

int send_and_process_ere(uint16_t ere[], uint16_t array_len) {
  HEADER_CLIENTD_PACKET   header;
  uint32_t                reserved_r_status;
  struct in6_addr        *req_relay_addr, *req_server_addr, rep_relay_addr, dns_addr;
  TYPE_LEN                tl;
  uint32_t                length_read;
  int                     i;
  char                    dn_suffix[128];
  char                    name[64];

  bzero((char *) &rep_relay_addr, sizeof(rep_relay_addr));
  
  if(!ssi_head) {
    warnx("No server to send request to even after solicit. Returning...\n");
    return 0;
  }

  req_relay_addr  = &ssi_head->relay_addr;
  req_server_addr = &ssi_head->server_addr;

  printf("Sending request to client daemon\n");
  i = request(NULL, req_relay_addr, req_server_addr, 0 /* C */, 0 /* R */,
	      0/* T */, 0/* transaction_id */, NULL, ere, array_len, NULL, 0);
  if(i == 0) {
    printf("Failure in getting Extension Request Extensions. Returning...\n");
    return 0;
  }

  length_read = 0;
  READ_DATA(&header, recv_msg, length_read, sizeof(header));
  header.pid        = ntohs(header.pid);
  header.request_no = ntohl(header.request_no);

  READ_DATA(&reserved_r_status, recv_msg, length_read, sizeof(reserved_r_status));
  reserved_r_status = ntohl(reserved_r_status);

  if(GET_BITS_MIDDLE(reserved_r_status, 8, 1)/* r bit */)
    READ_DATA(&rep_relay_addr, recv_msg, length_read, sizeof(rep_relay_addr));

  if(GET_BITS_LOW(reserved_r_status, 7)) {
    warnx("Non-zero status field: %u in the reply message\n",
	   GET_BITS_LOW(reserved_r_status, 7));
    return 0;
  }

  printf("length read %u len %u\n", length_read, n);
  while(length_read < n) {
    READ_DATA(&tl, recv_msg, length_read, sizeof(tl));
    tl.type      = ntohs(tl.type);
    tl.length    = ntohs(tl.length);
    switch(tl.type){
      
    case DOMAIN_NAME_SERVER_EXTENSION:
      {
	uint temp_length = 0;
	while(temp_length < tl.length) {
	  READ_DATA(&dns_addr, recv_msg, length_read, sizeof(dns_addr));
	  temp_length += sizeof(dns_addr);
	  printf("domain name server: %s\n",
		 inet_ntop(AF_INET6, &dns_addr, name, 64));
	}
      }
      break;
      
    case DOMAIN_NAME_SUFFIX_EXTENSION:
      READ_DATA(dn_suffix, recv_msg, length_read, tl.length);
      dn_suffix[tl.length] = '\0';
      printf("domain name suffix: %s\n", dn_suffix);
      break;

    default:
      /* Ignore the extension */
      printf("extension type %u\n", tl.type);
      length_read += tl.length;
      
    }/* switch */
  }

  return 1;
}

# define INIT                        1
# define ANY_ADDR                    2
# define ANY_ADDR_WITH_LIFETIME      3
# define SPECIFIC_ADDR               4
# define SPECIFIC_ADDR_WITH_LIFETIME 5
# define RELEASE                     6
# define DNS_UPDATE_NON_DHCP         7
# define ERE                         8
# define REQUEST_WITH_CLEAR          9
# define RELEASE_DNS_RECORDS         10

main(int argc, char *argv[]){
  struct in6_addr x_addr, ip_addr;
  char name[64], dns_name[128];
  uint8_t A, P, choice;
  uint32_t preferred_lifetime;
  uint32_t valid_lifetime;
  int16_t ere[128];
  struct in6_addr addr[100];
  int prefix_len;
  
  while(1) {
    printf("\nChoices: (1) init (2) any_addr (3) any_addr_with_lifetime\n\t"
	   " (4) specific_addr (5) specific_addr_with_lifetime (6) release\n\t"
	   " (7) DNS_UPDATE_NON_DHCP (8) ERE (9) REQUEST_WITH_CLEAR\n\t"
	   " (10) RELEASE_DNS_RECORDS \n");
    printf("Enter you choice: ");
    scanf("%d", &choice);

    switch(choice) {

    case INIT:
      send_and_process_init();
      print_server_solicit_info(ssi_head);
      break;

    case ANY_ADDR:
      send_and_process_request_any_addr(0, 1, "bhoot.com");
      print_leased_addr_list(liaddr_list_head);
      break;
      
    case ANY_ADDR_WITH_LIFETIME:
      printf("Enter preferred-lifetime: \n");
      scanf("%d", &preferred_lifetime);
      printf("Enter valid-lifetime: \n");
      scanf("%d", &valid_lifetime);
      send_and_process_request_any_addr_with_lifetime(preferred_lifetime, valid_lifetime,
						      0, 1, "bhoot.com");
      print_leased_addr_list(liaddr_list_head);
      break;
      
    case SPECIFIC_ADDR:
      printf("Enter specific addr: \n");
      scanf("%s", name);
      if(inet_pton(AF_INET6, name, &ip_addr) <= 0)
	errx(1, "Error converting address to binary\n");
      getchar();
      printf("Enter prefix-len: \n");
      scanf("%d", &preferred_lifetime);
      send_and_process_request_specific_addr(&ip_addr, 0, 1, "specific.org", prefix_len);
      print_leased_addr_list(liaddr_list_head);
      break;
 
    case SPECIFIC_ADDR_WITH_LIFETIME:
      printf("Enter specific addr: \n");
      scanf("%s", name);
      if(inet_pton(AF_INET6, name, &ip_addr) <= 0)
	errx(1, "Error converting address to binary\n");
      //getchar();
      printf("Enter preferred-lifetime: \n");
      scanf("%d", &preferred_lifetime);
      printf("Enter valid-lifetime: \n");
      scanf("%d", &valid_lifetime);
      send_and_process_request_specific_addr_with_lifetime(preferred_lifetime, valid_lifetime,
							   &ip_addr, 0, 1, "specific.org");
      print_leased_addr_list(liaddr_list_head);
      break;

    case RELEASE:
      {
	LIADDR *li;

	printf("Enter address: ");
	scanf("%s", name);
	getchar();
	if(inet_pton(AF_INET6, name, &ip_addr) <= 0)
	  errx(1, "Error converting address to binary\n");

	if(inet_pton(AF_INET6, X_ADDR, &x_addr) <= 0)
	  errx(1, "Error converting x-address to binary\n");

	for(li = liaddr_list_head; li; li = li->next) {
	  if(IN6_ARE_ADDR_EQUAL(&li->ip_addr, &ip_addr)) {
	    send_and_process_release(0, 0, 0, &li->ip_addr, 0/* R */,
				     &li->server_addr,
				     &x_addr);
	    print_leased_addr_list(liaddr_list_head);
	    break;
	  }
	}
	if(!li)
	  printf("Address %s couldn't be found\n", name);
      }
      break;

    case DNS_UPDATE_NON_DHCP:
	printf("Enter address: ");
	scanf("%s", name);
	getchar();
	if(inet_pton(AF_INET6, name, &ip_addr) <= 0)
	  errx(1, "Error converting address to binary\n");

	printf("Enter dns_name: ");
	scanf("%s", dns_name);
	getchar();

	send_and_process_dns_update(0, 1, &ip_addr, dns_name);

      break;

    case ERE:
      {
	int ere_length = 0;
	do {
	  printf("Enter extension no: \n");
	  scanf("%d", &ere[ere_length]);
	}
	while(ere[ere_length++] > 0);
	send_and_process_ere(ere, ere_length - 1);
      }
      break;

    case REQUEST_WITH_CLEAR:
      {
	LIADDR **temp_liaddr, *t_liaddr;
	int i = 0;

	printf("The current addr list is: ");
	for(t_liaddr = liaddr_list_head; t_liaddr; t_liaddr = t_liaddr->next) {
	  printf("%s ", inet_ntop(AF_INET6, &t_liaddr->ip_addr, name, 64));
	}
	printf("\nEnter the list of addresses you don't want to release. Stop by typing stop\n\n");

	while(1) {
	  printf("addr: ");
	  scanf("%s", name);
	  getchar();
	  if(strcasecmp(name, "stop") == 0 || strcasecmp(name, "exit") == 0)
	    break;

	  if(inet_pton(AF_INET6, name, &addr[i++]) <= 0)
	    errx(1, "Error converting address to binary\n");
	  
	}
	send_and_process_request_clear(addr, i);
	print_leased_addr_list(liaddr_list_head);
      }
      break;

    case RELEASE_DNS_RECORDS:
      {
	LIADDR *li;
	int A, P;
	
	printf("Enter address: ");
	scanf("%s", name);
	getchar();
	
	printf("Enter release A(0/1): ");
	scanf("%d", &A);
	getchar();

	printf("Enter release P(0/1): ");
	scanf("%d", &P);
	getchar();
	
	if(inet_pton(AF_INET6, name, &ip_addr) <= 0)
	  errx(1, "Error converting address to binary\n");

	if(inet_pton(AF_INET6, X_ADDR, &x_addr) <= 0)
	  errx(1, "Error converting x-address to binary\n");

	for(li = liaddr_list_head; li; li = li->next) {
	  if(IN6_ARE_ADDR_EQUAL(&li->ip_addr, &ip_addr)) {
	    send_and_process_release(A, P, 0, &li->ip_addr, 0/* R */,
				     &li->server_addr,
				     &x_addr);
	    print_leased_addr_list(liaddr_list_head);
	    break;
	  }
	}
	if(!li)
	  printf("Address %s couldn't be found\n", name);
      }
      break;

    dafault:
      printf("Bad choice: %d\n", choice);
    }
  }
   
}
