void create_solicit(FD_INIT_LIST *appl, INIT_DATA *data){
  SOLICIT sol;
  int     r;

  n = sizeof(sol);
  bzero((char *) &sol, n);
  
  sol.msg_type              = DHCP_SOLICIT;
  sol.c_p_reserved          = 0;
  sol.prefix_len_solicit_id = 0;

  SET_BITS_MIDDLE(sol.c_p_reserved, data->c, 8, 1); /* sol.c = data->c */
  SET_BITS_MIDDLE(sol.c_p_reserved, 1, 7, 1);      /* sol.p = 1 Let me obtain prefixes for others. 
						      You may use "solicit.p" instead */

  if(appl->interface->solicit_id > (1 << SOLICIT_ID_LEN) ||
     appl->interface->solicit_id == ((1 << SOLICIT_ID_LEN) - 1))
    appl->interface->solicit_id = 0;

  if(data->a == 1){
    
    appl->sol_xact_id = /* sol.solicit_id = */ ++(appl->interface->solicit_id);
    SET_BITS_LOW(sol.prefix_len_solicit_id, appl->sol_xact_id, 9);
  }
  else{
    appl->sol_xact_id = /* sol.solicit_id = */ 0;
    SET_BITS_LOW(sol.prefix_len_solicit_id, 0, 9);
    /* 
     * Note: We let the solicit id of interface remain the older one so that 
     * next time we want list of all servers solicit_id is initial++ not 1
     * (initial != 0)
     */
  }

  sol.client_link_local_addr = *(appl->interface->link_local_addr);

#ifdef DEBUG
  printf("Creating solicit to server...\n");
  printf("msg_type = %u c = %u p = %u res = %u prefix_len = %u solicit_id = %u\n",
	 sol.msg_type,
	 GET_BITS_MIDDLE(sol.c_p_reserved, 8, 1) /* sol.c        */,
	 GET_BITS_MIDDLE(sol.c_p_reserved, 7, 1) /* sol.p        */,
	 GET_BITS_LOW(sol.c_p_reserved, 6)       /* sol.reserved */,
	 GET_BITS_HIGH(sol.prefix_len_solicit_id, 7, 16),
	 GET_BITS_LOW(sol.prefix_len_solicit_id, 9));

  printf("client's link local address = %s\n", 
	 inet_ntop(AF_INET6, &sol.client_link_local_addr, name, 127));
  printf("relay's address = %s\n", inet_ntop(AF_INET6, &sol.relay_addr, name, 127));
#endif

  /* Convert to n/w byte order */
  sol.prefix_len_solicit_id = htons(sol.prefix_len_solicit_id);

  /* 
   * Note: I am using the same sockfd opened on CLIENT PORT for
   * sending outgoing packets. If you want to use different port
   * (for may be firewalls) just open a new UDP socket, bind it
   * to link local addr and send on it. But remember to close it.
   * This is not very good as you end up opening a new socket each
   * time (lots of binds). If you want to save on this you can add
   * a new field to FD_INFO struct in INTERFACE struct.
   */

  memcpy(appl->msg, &sol, n);
  appl->msg_len                = n;
  appl->serv_sockaddr.sin6_addr = server_multicast_addr;//in6addr_any;//in6addr_loopback;//
  //if(inet_pton(AF_INET6, "3ffe:305:1002:1:2b0:d0ff:fe5c:4aee", &appl->serv_sockaddr.sin6_addr) <= 0)
  //errx(1, "Can't convert address to binary. Exiting...\n");
  appl->serv_sockaddr.sin6_port = htons(AGENT_PORT);
  /*
   * Note that when you retransmit you will find no of attempts == 0.
   * Then you must set "when_sent"
   */
  appl->no_of_attempts  = 0;
  appl->request_pending = 1;
  appl->adv_received    = 0;
  appl->timeout         = ADV_MSG_TIMEOUT;

  /* Note that in the following I haven't used
   * add_in_sec(&current_time,
   *            (MIN_SOL_DELAY + (random() % (MAX_SOL_DELAY - MIN_SOL_DELAY + 1))),
   *            &temp_init->next_transmission
   *            );
   *
   * because it gives a range in quantum of seconds not millisecs.
   */

  /* Note 1'st solicit is to be delayed between MIN & MAX_SOL_DELAY */
  printf("\n...current_time = %u %u...\n", (current_time).tv_sec, (current_time).tv_usec);
  //r = 1;
  /* Imp: Uncomment the following comment and comment the above line: r = 1*/
  r = (MIN_SOL_DELAY * 1000 + (random() % (((MAX_SOL_DELAY - MIN_SOL_DELAY) * 1000) + 1) ));
  add_in_ms(&current_time,
	    r,
	    &appl->next_transmission
	    );

  printf("\n...random = %d Next transmission = %u %u...\n",
	 r,
	 (appl->next_transmission).tv_sec,
	 (appl->next_transmission).tv_usec);
  return;
}
