void print(INTERFACE **head){
  /*
   * Prints info about those entries in link list which have a
   * link local address. Thus these are the interfaces we would
   * like to configure.
   * This function also removes unwanted interfaces from the
   * list and opens a socket for each address. It also binds
   * these sockets.
   * In addition the bits of fd_set f are also set to include
   * the fds we are opening for each IPv6 address.
   */
  IADDR     *tmp_addr;
  INTERFACE *tmp_interface = *head;
  char buf[128];
  uint8_t i;
  FD_INFO *temp_fd_info, **temp_fd;

  while(tmp_interface != NULL){
    if(tmp_interface->link_local_addr != NULL){
      /*
       * This is the type of interface we look for (the one
       * with a link local address).
       */

#ifdef PRINT
      printf("Link local addr of [%u] name <%s> = %s Link layer address = ",
	     tmp_interface->index,
	     tmp_interface->name,
	     inet_ntop(AF_INET6, tmp_interface->link_local_addr, buf, sizeof(buf)));

      for(i = 0; i < 5; i++)
	printf("%02x:",tmp_interface->hardware_addr.hw_addr[i]);
      printf("%02x\n",tmp_interface->hardware_addr.hw_addr[i]);
#endif
      
      tmp_interface->solicit_id     = 1024;                                    /* 1024 means no solicit done           */
      tmp_interface->xact_id        = MIN_CLIENT_TRANSACTION_ID;
      tmp_interface->link_local_fd  = (FD_INFO *)malloc(sizeof(FD_INFO));
      temp_fd_info                  = tmp_interface->link_local_fd;
      temp_fd_info->next            = NULL;
      temp_fd_info->i_addr          = NULL;//tmp_interface->link_local_addr;//
      temp_fd_info->fd              = socket(AF_INET6, SOCK_DGRAM, 0);
      if(temp_fd_info->fd < 0)
	err_dump("Error in opening datagram socket! Exiting...");

#ifdef PRINT
      printf("fd = %d addr = %s\n", temp_fd_info->fd, 
	     inet_ntop(AF_INET6, tmp_interface->link_local_addr, buf, sizeof(buf)));
#endif

      FD_SET(temp_fd_info->fd, &f);
      nfds = (nfds > temp_fd_info->fd) ? nfds : temp_fd_info->fd;
      bzero((char *) &temp_fd_info->sockaddr, sizeof(temp_fd_info->sockaddr));
      
#ifdef SIN6_LEN
      temp_fd_info->sockaddr.sin6_len      = sizeof(temp_fd_info->sockaddr);    /* 4.3 BSD doesn't have this field   */
#endif
      
      temp_fd_info->sockaddr.sin6_family   = AF_INET6;                          /* Family as required for Ipv6       */
      temp_fd_info->sockaddr.sin6_port     = htons(CLIENT_PORT);                /* The CLD talks on CLIENT_PORT      */
      temp_fd_info->sockaddr.sin6_flowinfo = 0;                                 /* Flow & Traffic Class:  unimportant*/
      temp_fd_info->sockaddr.sin6_addr     = *(tmp_interface->link_local_addr); /* CLIENT_ADDRESS                    */
      temp_fd_info->sockaddr.sin6_scope_id = tmp_interface->index;              /* Note: interface index goes here   */
      
      if(setsockopt(temp_fd_info->fd,
		    IPPROTO_IPV6, 
		    IPV6_MULTICAST_IF, 
		    (char *) &tmp_interface->index,
		    sizeof(tmp_interface->index)
		    ) == -1){
	err_dump("Couldn't set socket option to IPV6_MULTICAST_IF. Exiting...");
      }
      
      if(bind(temp_fd_info->fd, (struct sockaddr *) &temp_fd_info->sockaddr,
	      sizeof(temp_fd_info->sockaddr)) < 0){
	warnx("Error while binding ll address of interface %d ! Exiting... Address = ",
	       tmp_interface->index);
	printf("%s\n",inet_ntop(AF_INET6, tmp_interface->link_local_addr, buf, sizeof(buf)));
      }


      tmp_addr = tmp_interface->addr_list_head;
      temp_fd  = & tmp_interface->ip_addr_fd_head;

#ifdef PRINT
      printf("\t\tIPv6 addr list ");
#endif

      while(tmp_addr != NULL){

#ifdef PRINT
	printf("%s ",inet_ntop(AF_INET6, & tmp_addr->ip_addr, buf, sizeof(buf)));
#endif

	*temp_fd                    = (FD_INFO *)malloc(sizeof(FD_INFO));
	(*temp_fd)->next            = NULL;
	(*temp_fd)->i_addr          = tmp_addr;
	(*temp_fd)->fd              = socket(AF_INET6, SOCK_DGRAM, 0);
	if((*temp_fd)->fd < 0)
	  err_dump("Error in opening datagram socket! Exiting...");
	
#ifdef PRINT
	printf("\nfd = %d addr = %s\n", (*temp_fd)->fd, 
	       inet_ntop(AF_INET6, & tmp_addr->ip_addr, buf, sizeof(buf)));
#endif
	
	FD_SET((*temp_fd)->fd, &f);
	nfds = (nfds > (*temp_fd)->fd) ? nfds : (*temp_fd)->fd;
	
	bzero((char *) &(*temp_fd)->sockaddr, sizeof((*temp_fd)->sockaddr));
	
#ifdef SIN6_LEN
	(*temp_fd)->sockaddr.sin6_len      = sizeof((*temp_fd)->sockaddr);    /* 4.3 BSD doesn't have this field   */
#endif
	
	(*temp_fd)->sockaddr.sin6_family   = AF_INET6;                          /* Family as required for Ipv6       */
	(*temp_fd)->sockaddr.sin6_port     = htons(CLIENT_PORT);                /* The CLD talks on CLIENT_PORT      */
	(*temp_fd)->sockaddr.sin6_flowinfo = 0;                                 /* Flow & Traffic Class:  unimportant*/
	(*temp_fd)->sockaddr.sin6_addr     = tmp_addr->ip_addr;                 /* CLIENT_ADDRESS                    */
	(*temp_fd)->sockaddr.sin6_scope_id = 0;                                 /* Not really required.              */
	
	if(bind((*temp_fd)->fd, (struct sockaddr *) &(*temp_fd)->sockaddr,
		sizeof((*temp_fd)->sockaddr)) < 0){
	  printf("Error while binding address of interface %d ! Address = ", tmp_interface->index);
	  printf("%s\n",inet_ntop(AF_INET6, & tmp_addr->ip_addr, buf, sizeof(buf)));
	}

	tmp_addr = tmp_addr->next;
	temp_fd  = & (*temp_fd)->next;
      }
#ifdef PRINT
      printf("\n");
#endif
    }
    else{
      /*
       * Delete this interface
       */
      if(tmp_interface->prev != NULL)
	tmp_interface->prev->next = tmp_interface->next; 

      if(tmp_interface->next != NULL)
	tmp_interface->next->prev = tmp_interface->prev;
      
      if(tmp_interface == *head)
	*head = (*head)->next;
      
      free(tmp_interface);
    }
    tmp_interface = tmp_interface->next;
  }
}

/*********************************************************************************/

# define ADDRESS_FAMILY AF_INET6
/*********************************************************************************/

int                mib[] = { CTL_NET, PF_ROUTE, 0, ADDRESS_FAMILY, NET_RT_IFLIST, 0 };
caddr_t            buf;
struct if_msghdr  *ifm;
struct ifa_msghdr *ifam;
struct sockaddr   *sa;
INTERFACE         *temp_interface;

/*********************************************************************************/
void discover_interfaces(){
  /*
   * Discovers all IPv6 interfaces and binds each
   * available address to CLIENT_PORT by opening
   * a socket per address per interface including
   * link local address.
   */
  caddr_t cp;
  int     interface_in_list;
  int     size;
  
  if (sysctl(mib, 6, NULL, &size, NULL, 0) < 0) {
    perror("sysctl0");
    exit(1);
  }

  if ((buf = malloc(size)) == NULL) {
    fprintf(stderr, "malloc failed\n");
    exit(1);
  }

  if (sysctl(mib, 6, buf, &size, NULL, 0) < 0) {
    perror("sysctl");
    exit(1);
  }

  for (cp = buf; cp < buf + size; cp += ifm->ifm_msglen) {
    ifm = (struct if_msghdr *)cp;
    ifam = (struct ifa_msghdr *)cp;
    interface_in_list = 1;
    temp_interface = find_interface(ifm->ifm_index, in_list_head);

    if(temp_interface == NULL){
      temp_interface = (INTERFACE *) malloc(sizeof(INTERFACE));
      bzero(temp_interface, sizeof(*temp_interface));
      temp_interface->index = ifm->ifm_index;
      interface_in_list = 0;
    }
    
    if (ifm->ifm_type == RTM_IFINFO)
      ifinfo();
    else
      ifaddr();

    if(interface_in_list == 0){
      insert_in_list((IADDR *)temp_interface, (IADDR **)&in_list_head);
      if(in_list_head == NULL)
	printf("Null after inserting");
    }
  }

  free(buf);  
  print(&in_list_head);
}
/*********************************************************************************/

#define ROUNDUP(a) \
        ((a) > 0 ? (1 + (((a) - 1) | (sizeof(long) - 1))) : sizeof(long))
#define ADVANCE(x, n) (x += ROUNDUP((n)->sa_len))
/*********************************************************************************/

ifinfo(){
  caddr_t cp;
  int len, i;
  
  cp = (caddr_t)(ifm + 1);
  len = ifm->ifm_msglen;
  
  for (i = 0; i < RTAX_MAX; i++)
    if (ifm->ifm_addrs & (1 << i)) {
      sa = (struct sockaddr *)cp;
      bit(i);
      sock();
      ADVANCE(cp, sa);
    }
  
  if (cp != (caddr_t)ifm + len)
    printf("delta = %d\n", cp - ((caddr_t)ifm + len));
}
/*********************************************************************************/

ifaddr(){
  caddr_t cp;
  int len, i;
  
  cp = (caddr_t)(ifam + 1);
  len = ifam->ifam_msglen;
  
  for (i = 0; i < RTAX_MAX; i++)
    if (ifam->ifam_addrs & (1 << i)) {
      sa = (struct sockaddr *)cp;
      bit(i);
      sock();
      ADVANCE(cp, sa);
    }
  
  if (cp != (caddr_t)ifam + len)
    printf("delta = %d\n", cp - ((caddr_t)ifam + len));
}
/*********************************************************************************/

bit(int i){
  switch (i) {
  case RTAX_DST:
    return;
    
  case RTAX_GATEWAY:
    return;
    
  case RTAX_NETMASK:
    return;
    
  case RTAX_GENMASK:
    return;
    
  case RTAX_IFP:
    return;
    
  case RTAX_IFA:
    return;
    
  case RTAX_AUTHOR:
    return;
    
  case RTAX_BRD:
    return;
    
  default:
    return;
  }
}
/*********************************************************************************/

sock(){
  u_char *cp;
  char buf[64];
  int i;
  IADDR *temp_lease;
  
  switch (sa->sa_family) {

  case AF_UNSPEC:
    break;
    
  case AF_INET:
    return;
    
  case AF_INET6:
#define SIN6(x)	((struct sockaddr_in6 *)x)
    temp_lease            = (IADDR *)malloc(sizeof(IADDR));
    memset((char *) temp_lease, 0, sizeof(*temp_lease));
    temp_lease->ip_addr   = SIN6(sa)->sin6_addr;
    temp_lease->interface = temp_interface;
    
    if(IN6_IS_ADDR_LOOPBACK(& temp_lease->ip_addr)){
      temp_interface->loopback_addr    = (struct in6_addr *)malloc(sizeof(struct in6_addr));
      *temp_interface->loopback_addr   = temp_lease->ip_addr;
      free(temp_lease);
      
    }
    else{
      if(IN6_IS_ADDR_LINKLOCAL(& temp_lease->ip_addr)){
	temp_interface->link_local_addr  = (struct in6_addr *)malloc(sizeof(struct in6_addr));
	*temp_interface->link_local_addr = temp_lease->ip_addr;
	free(temp_lease);
      }
      else
	insert_in_list(temp_lease, & temp_interface->addr_list_head);
    }
    return;
    
  case AF_LINK:
#define SDL(x)	((struct sockaddr_dl *)x)
   
    sprintf(temp_interface->name,"%*s",
	    SDL(sa)->sdl_nlen,
	    SDL(sa)->sdl_data);
    
    cp = (u_char *)(SDL(sa)->sdl_data + SDL(sa)->sdl_nlen);
    
    for (i = 0; i < SDL(sa)->sdl_alen; i++){
      temp_interface->hardware_addr.hw_addr[i] = *cp++;
    }
    return;
    
  default:
  }
}
