void handle_init(char *recv_msg,
		int n,
		FD_ACCEPT_LIST *accept_prev,
		FD_ACCEPT_LIST *temp_accept){
  /*
   * Check if message is a valid init. Note down appropriate
   * parameters of init e.x. want_to_reconfig, c, p.
   */
  HEADER_CLIENTD_PACKET  header;
  SOLICIT_PARAMS         solicit;
  INTERFACE             *interface;
  INIT_DATA              init_data;
  uint8_t                s1, s2, i;
  FD_INIT_LIST          **temp_fil = &init_list_head, *t_fil;

  s1 = sizeof(header);
  s2 = sizeof(solicit);

  if(n != (s1 + s2)){
    printf("Invalid sized init message from appl\n");

    /* Send error back to appl/lib */
    if(n > s1)
      memcpy(& header,  recv_msg, s1);
    else
      bzero(&header, s1);

    header.error = INVALID_MESSAGE_FORMAT;
    if(send(temp_accept->fd, &header, s1, 0) != s1)
      err_dump("Error while sending to appl\n");

    return;
  }

  memcpy(& header,  recv_msg, s1);
  memcpy(& solicit, recv_msg + s1, s2);

  if(header.msg_type != 0){
    printf("Invalid msg_type of init message\n");

    /* Send error back to appl/lib */
    header.error = INVALID_MESSAGE_FORMAT;
    if(send(temp_accept->fd, &header, s1, 0) != s1)
      err_dump("Error while sending to appl\n");

    return;
  }
  /* You can check here if pid is already present. If yes delete the appl as it
   * must not send init twice
   */
  /* Convert to host byte order */
  solicit.if_index     = ntohl(solicit.if_index);
  solicit.res_c_p_a_w  = ntohl(solicit.res_c_p_a_w);

  interface         = find_interface(solicit.if_index, in_list_head);
  if(interface == NULL){
    if(solicit.if_index == 0){
      interface = get_main_interface();
    }
    else{
      printf("Invalid interface specified in init message\n");
      
      /* Send error back to appl/lib */
      header.error = MESSAGE_FIELD_INVALID;
      if(send(temp_accept->fd, &header, s1, 0) != s1)
	err_dump("Error while sending to appl\n");
      
      return;
    }
  }

  /* Convert to host byte order */
  header.pid                  = ntohs(header.pid);
  header.request_no           = ntohl(header.request_no);

  init_data.c                 = GET_BITS_MIDDLE(solicit.res_c_p_a_w, 4, 1);/* solicit.c */;
  init_data.p                 = GET_BITS_MIDDLE(solicit.res_c_p_a_w, 3, 1);/* solicit.p */;
  init_data.a                 = GET_BITS_MIDDLE(solicit.res_c_p_a_w, 2, 1);/* solicit.a */;

  temp_init                   = (FD_INIT_LIST *)malloc(sizeof(FD_INIT_LIST));
  bzero((char *)temp_init, sizeof(*temp_init));

  temp_init->if_fd            = interface->link_local_fd;
  //temp_init->ip_addr          = interface->link_local_addr;
  temp_init->interface        = interface;
  temp_init->tcp_fd           = temp_accept->fd;
  temp_init->msg              = malloc(MAX_DHCP_MSG);
  temp_init->serv_sockaddr    = server_sockaddr;
  temp_init->serv_sockaddr.sin6_port = htons(AGENT_PORT);
  temp_init->msg_len          = 0;
  temp_init->timeout          = ADV_MSG_TIMEOUT;
  temp_init->msg_type         = header.msg_type;
  temp_init->pid              = header.pid;
  temp_init->request_no       = header.request_no;
  memcpy(&temp_init->msg_data[0], &init_data, sizeof(init_data));
  temp_init->no_of_attempts   = 0;
  temp_init->want_to_reconfig = GET_BITS_MIDDLE(solicit.res_c_p_a_w, 1, 1);/* solicit.w */;
  temp_init->request_pending  = 1;
  temp_init->adv_received     = 0;

  /* Now delete temp_accept from accept list and add to init list */
  if(accept_prev != NULL)
    accept_prev->next = temp_accept->next;
  else
    fd_accept_list_head = temp_accept->next;

  free(temp_accept);
  insert((FD_ACCEPT_LIST *)temp_init, (FD_ACCEPT_LIST **)&fd_init_list_head);

  /*
   * Create and send out solicit message
   * Note: Two appl can use same sol_xact_id = 0 if solicit.a == 0. Otherwise not.
   * Therefore when server replies back with solicit id 0 you have to check entire
   * list to reply to appls not till first matching entry.
   */
  i = try_cached_adv(temp_init, &init_data);

  if(i == SUCCESS){

#ifdef DEBUG
    printf("Successful in using cached advertise info\n");
#endif
    return;
  }

  if(i == SEND_ERROR){

#ifdef DEBUG
    printf("Would not try resending.\n ");
#endif
    return;
  }

  create_solicit(temp_init, &init_data);
  return;
}
