#ifndef DEBUG
#define DEBUG
#endif

void handle_request_from_appl(char *msg, int len, FD_INIT_LIST *appl){
  HEADER_CLIENTD_PACKET header;
  uint32_t              length_read;
  int                   result;
  
  if(appl->request_pending){
    printf("Can't handle a new request from appl while a previous one is pending\n");

#ifdef DEBUG
    printf("Two consecutive requests from appl with pid = %u\n", appl->pid);
#endif

    return;
  }
  if(len < sizeof(header)){
    printf("Length of packet from appl < minimum(sizeof header) = %u\n", sizeof(header));
    return;
  }

  memcpy(&header, msg, sizeof(header));
  length_read          = sizeof(header);

  header.pid           = ntohs(header.pid);
  header.request_no    = ntohl(header.request_no);

  appl->msg_type       = header.msg_type;
  appl->pid            = header.pid;/* If the appl passed the descriptor to some
				     other process pid would have changed */
  appl->request_no     = header.request_no;
  appl->no_of_attempts = 0;

  switch(header.msg_type){

  case DHCP_SOLICIT:
    {
      uint32_t     res_c_p_a_w;
      SOLICIT_DATA sol_data;

      READ_DATA(&res_c_p_a_w, msg, length_read, sizeof(res_c_p_a_w));
      res_c_p_a_w = ntohl(res_c_p_a_w);

      sol_data.c  = GET_BITS_MIDDLE(res_c_p_a_w, 4, 1);/* solicit.c */;
      sol_data.p  = GET_BITS_MIDDLE(res_c_p_a_w, 3, 1);/* solicit.p */;
      sol_data.a  = GET_BITS_MIDDLE(res_c_p_a_w, 2, 1);/* solicit.a */;
      memcpy(&appl->msg_data[0], &sol_data, sizeof(sol_data));

      result = try_cached_adv(appl, &sol_data);

      if(result == SUCCESS){
#ifdef DEBUG
	printf("Successful in using cached advertise info\n");
#endif
	return;
      }
     
      if(result == SEND_ERROR){	
#ifdef DEBUG
	printf("Would not try resending.\n ");
#endif
	return;
      }
      
      create_solicit(appl, &sol_data);
    }
    break;


  case DHCP_REQUEST:
    {
      struct{
	uint16_t res_u_c_r_t;
	uint16_t transaction_id;
      }s;

      struct in6_addr unicast_addr;
      REQUEST req;
      uint8_t u = 0;
      FD_INFO *temp_fd_info;

      bzero((char *) &req, sizeof(req));
      req.msg_type     = DHCP_REQUEST;
      req.c_r_reserved = 0;

      READ_DATA(&s, msg, length_read, sizeof(s));
      s.res_u_c_r_t    = ntohs(s.res_u_c_r_t);
      s.transaction_id = ntohs(s.transaction_id);
      
      if(GET_BITS_MIDDLE(s.res_u_c_r_t, 4, 1)/* U bit */){
	/* This means client's unicast address is present */
	u = 1;
	READ_DATA(&unicast_addr, msg, length_read, sizeof(unicast_addr));
      }

      READ_DATA(&req.relay_addr, msg, length_read, sizeof(req.relay_addr));
      READ_DATA(&req.server_addr, msg, length_read, sizeof(req.server_addr));

      if(//Imp: IN6_IS_ADDR_UNSPECIFIED(&req.relay_addr) ||
	 IN6_IS_ADDR_LOOPBACK(&req.relay_addr) ||
	 IN6_IS_ADDR_MULTICAST(&req.relay_addr)){
	/* Relay address must be unicast addr */
#ifdef DEBUG
	printf("Invalid relay address in request message from appl\n");
#endif
	send_error(appl, MESSAGE_FIELD_INVALID);
	return;
      }

      if(IN6_IS_ADDR_UNSPECIFIED(&req.server_addr)){
#ifdef DEBUG
	printf("Appl sent unspecified server addr in request. Ignoring the request\n");
#endif
	send_error(appl, MESSAGE_FIELD_INVALID);
	return;
      }

      if(IN6_IS_ADDR_LOOPBACK(&req.server_addr)){
#ifdef DEBUG
	printf("Appl sent loopback server addr in request. Ignoring the request\n");
#endif
	send_error(appl, MESSAGE_FIELD_INVALID);
	return;
      }
      if(IN6_IS_ADDR_MULTICAST(&req.server_addr)){
	/* If the client already has an IP address of sufficient scope to
	 * directly reach the server, then the client SHOULD unicast the Request
	 * to the server.  Otherwise, if the server is off-link, the client
	 * unicasts the Request message to the appropriate relay.
	 */
#ifdef DEBUG
	printf("Request messages should not be multicast\n");
#endif
	send_error(appl, MESSAGE_FIELD_INVALID);
	return;
      }
      

      if(u){

	if(IN6_IS_ADDR_UNSPECIFIED(&unicast_addr)){
#ifdef DEBUG
	  printf("Appl sent unspecified source addr in request. Ignoring the request\n");
#endif
	  send_error(appl, MESSAGE_FIELD_INVALID);
	  return;
	}
	
	if(IN6_IS_ADDR_LOOPBACK(&unicast_addr)){
#ifdef DEBUG
	  printf("Appl sent loopback source addr in request. Ignoring the request\n");
#endif
	  send_error(appl, MESSAGE_FIELD_INVALID);
	  return;
	}

	if(IN6_IS_ADDR_MULTICAST(&unicast_addr)){
#ifdef DEBUG
	  printf("Appl sent multicast source addr in request. Ignoring the request\n");
#endif
	  send_error(appl, MESSAGE_FIELD_INVALID);
	  return;
	}
	/*
	 * Check whether the unicast address specified is of sufficient scope
	 * to reach the server and is a valid address for this interface
	 */
	temp_fd_info = find_address_info(&unicast_addr, appl->interface);
	if(!temp_fd_info){
#ifdef DEBUG
	  printf("Appl sent unknown source addr in request. Ignoring the request\n");
#endif
	  send_error(appl, MESSAGE_FIELD_INVALID);
	  return;
	}

	if(!IN6_IS_ADDR_LINKLOCAL(&req.server_addr)){
	  /* 
	   * Server if off link.
	   * Otherwise No problem. Unicast_addr would be of sufficient scope
	   */
	  if(IN6_IS_ADDR_LINKLOCAL(&unicast_addr)){
#ifdef DEBUG
	    printf("Address specified by appl is link local\n");
#endif
	    send_error(appl, MESSAGE_FIELD_INVALID);
	    return;
	  }
	    
	  if(!IN6_IS_ADDR_SITELOCAL(&req.server_addr) && IN6_IS_ADDR_SITELOCAL(&unicast_addr)){
	    /*
	     * Server is off link and not sitelocal => global
	     * Unicast addr is not ll but              sitelocal
	     * This means unicast addr is of insufficient scope
	     */
#ifdef DEBUG
	    printf("Address specified by appl is site local which is of insufficient scope\n");
#endif
	    send_error(appl, MESSAGE_FIELD_INVALID);
	    return;
	  }
	}
	/*
	 * If you are here it means you have an address of sufficient scope.
	 * You can unicast the request msg.
	 */
	appl->serv_sockaddr.sin6_addr = req.server_addr;
      }
      else{
	temp_fd_info = appl->interface->link_local_fd;
	/* If U bit is not set you must use relay if server if offlink */
	if(!IN6_IS_ADDR_UNSPECIFIED(&req.relay_addr))
	  appl->serv_sockaddr.sin6_addr = req.relay_addr;
	else
	  appl->serv_sockaddr.sin6_addr = req.server_addr;
      }
      
      appl->if_fd = temp_fd_info;
      //appl->ip_addr = &(appl->if_fd->i_addr->ip_addr);
      if(GET_BITS_LOW(s.res_u_c_r_t, 1)/* T bit */){
	appl->sol_xact_id  = s.transaction_id;
#ifndef JUNK
	printf("T: %u %u\n", appl->sol_xact_id, htons(appl->sol_xact_id));
#endif
	req.transaction_id = htons(appl->sol_xact_id);
      }
      else{
	if(appl->interface->xact_id >= MAX_CLIENT_TRANSACTION_ID)
	  appl->interface->xact_id = MIN_CLIENT_TRANSACTION_ID;

	appl->sol_xact_id  = appl->interface->xact_id++;
#ifndef JUNK
	printf("No T: %u %u\n", appl->sol_xact_id, htons(appl->sol_xact_id));
#endif
	req.transaction_id = htons(appl->sol_xact_id);
      }

      SET_BITS_HIGH(req.c_r_reserved, GET_BITS_MIDDLE(s.res_u_c_r_t, 3, 1)/* c bit */, 1, 8);
      SET_BITS_MIDDLE(req.c_r_reserved, GET_BITS_MIDDLE(s.res_u_c_r_t, 2, 1)/* r bit */, 7, 1);

      req.client_link_local_addr = *(appl->interface->link_local_addr);

      memcpy(appl->msg, &req, sizeof(req));
      /* Copy extensions from the application's request */
      memcpy(&appl->msg[sizeof(req)], msg + length_read, len - length_read);
      appl->msg_len                = sizeof(req) + len - length_read;
      appl->request_pending        = 1;

      appl->when_sent = current_time;
      appl->timeout   = REP_MSG_TIMEOUT;
      add_in_ms(&current_time,
		appl->timeout,
		&appl->next_transmission
		);
      
      send_msg_for_appl(appl);
    }
    break;


  case DHCP_RELEASE:
    {
      RELEASE   rel;
      uint32_t  res_r, start_of_ext;
      FD_INFO  *temp_fd_info;
      TYPE_LEN tl;

      bzero((char *) &rel, sizeof(rel));
      rel.msg_type   = DHCP_RELEASE;
      rel.r_reserved = 0;

      if(appl->interface->xact_id == MAX_CLIENT_TRANSACTION_ID)
	appl->interface->xact_id = MIN_CLIENT_TRANSACTION_ID;

      appl->sol_xact_id          = appl->interface->xact_id++;
      rel.transaction_id         = htons(appl->sol_xact_id);
      rel.client_link_local_addr = *(appl->interface->link_local_addr);

      READ_DATA(&res_r, msg, length_read, sizeof(res_r));
      res_r = ntohl(res_r);

      READ_DATA(&rel.server_addr, msg, length_read, sizeof(rel.server_addr));
      READ_DATA(&rel.x_addr, msg, length_read, sizeof(rel.x_addr));


      if(IN6_IS_ADDR_UNSPECIFIED(&rel.server_addr)){
#ifdef DEBUG
	printf("Appl sent unspecified server addr in release. Ignoring the release\n");
#endif
	send_error(appl, MESSAGE_FIELD_INVALID);
	return;
      }
      
      if(IN6_IS_ADDR_LOOPBACK(&rel.server_addr)){
#ifdef DEBUG
	printf("Appl sent loopback server addr in release. Ignoring the release\n");
#endif
	send_error(appl, MESSAGE_FIELD_INVALID);
	return;
      }
      if(IN6_IS_ADDR_MULTICAST(&rel.server_addr)){
	/* If the ``R'' bit is set, then the client MUST unicast the Release
	 * to the relay indicated in the ``X-address'' field.  Otherwise, the
	 * client unicasts the Release message directly to the server indicated
	 * in the ``server-address'' field.
	 */
#ifdef DEBUG
	printf("Release messages should not be multicast\n");
#endif
	send_error(appl, MESSAGE_FIELD_INVALID);
	return;
      }
      
      if(GET_BITS_LOW(res_r, 1)){
	/* If r_bit is set x_addr must be address of relay */
	/* Check if it is not unspecified or loopback or multicast */
	if(IN6_IS_ADDR_UNSPECIFIED(&rel.x_addr) ||
	   IN6_IS_ADDR_LOOPBACK(&rel.x_addr) ||
	   IN6_IS_ADDR_MULTICAST(&rel.x_addr)){
	  /* Relay address must be unicast addr */
#ifdef DEBUG
	  printf("Invalid relay address in release message from appl\n");
#endif
	  send_error(appl, MESSAGE_FIELD_INVALID);
	  return;
	}
	appl->serv_sockaddr.sin6_addr = rel.x_addr;
	appl->if_fd                   = appl->interface->link_local_fd;
      }
      else{
	if(IN6_IS_ADDR_UNSPECIFIED(&rel.x_addr)){
	  /* Note that you don't need this check as well as the next one
	   * because of the check "temp_fd_info = find_... ; if(!temp_fd_info)...
	   */
#ifdef DEBUG
	  printf("Appl sent unspecified source addr in release. Ignoring the release\n");
#endif
	  send_error(appl, MESSAGE_FIELD_INVALID);
	  return;
	}
	
	if(IN6_IS_ADDR_LOOPBACK(&rel.x_addr)){
#ifdef DEBUG
	  printf("Appl sent loopback source addr in release. Ignoring the release\n");
#endif
	  send_error(appl, MESSAGE_FIELD_INVALID);
	  return;
	}

	if(IN6_IS_ADDR_MULTICAST(&rel.x_addr)){
#ifdef DEBUG
	  printf("Appl sent multicast source addr in release. Ignoring the release\n");
#endif
	  send_error(appl, MESSAGE_FIELD_INVALID);
	  return;
	}
	/* Check if x_addr is of sufficient scope to reach the server     */
	/* But first check whether this address belongs to this interface */
	temp_fd_info = find_address_info(&rel.x_addr, appl->interface);
	if(!temp_fd_info){
#ifdef DEBUG
	  printf("Appl sent unknown source addr in release. Ignoring the release\n");
#endif
	  send_error(appl, MESSAGE_FIELD_INVALID);
	  return;
	}
	if(!IN6_IS_ADDR_LINKLOCAL(&rel.server_addr)){
	  /* 
	   * Server if off link.
	   * Otherwise No problem. Unicast_addr would be of sufficient scope
	   */
	  if(IN6_IS_ADDR_LINKLOCAL(&rel.x_addr)){
#ifdef DEBUG
	    printf("Address specified by appl is link local\n");
#endif
	    send_error(appl, MESSAGE_FIELD_INVALID);
	    return;
	  }
	    
	  if(!IN6_IS_ADDR_SITELOCAL(&rel.server_addr) && IN6_IS_ADDR_SITELOCAL(&rel.x_addr)){
	    /*
	     * Server is off link and not sitelocal => global
	     * Unicast addr is not ll but              sitelocal
	     * This means unicast addr is of insufficient scope
	     */
#ifdef DEBUG
	    printf("Address specified by appl is site local which is of insufficient scope\n");
#endif
	    send_error(appl, MESSAGE_FIELD_INVALID);
	    return;
	  }
	}
	appl->serv_sockaddr.sin6_addr = rel.server_addr;
	appl->if_fd                   = temp_fd_info;
      }
      /* process extensions to find out what has appl released */
      /* If no extensions are specified all resources are released */
      /* Check whether address relesed == x_addr. if so, error */
      start_of_ext = length_read;

      for(;length_read < len;){
	
	READ_DATA(&tl, msg, length_read, sizeof(tl));
	tl.type      = ntohs(tl.type);
	tl.length    = ntohs(tl.length);
	
	switch(tl.type){
	  
	case IP_ADDRESS_EXTENSION:
	  {
	    struct{
	      uint32_t status     : 8;
	      uint32_t C          : 1;
	      uint32_t I          : 1;
	      uint32_t L          : 1;
	      uint32_t Q          : 1;
	      uint32_t A          : 1;
	      uint32_t P          : 1;
	      uint32_t reserved   : 8;
	      uint32_t scope      : 3;
	      uint32_t prefix_len : 7;
	    } s;
	    
	    uint32_t         temp_ext;
	    struct in6_addr  ip_addr;
	    uint32_t         preferred_lifetime, valid_lifetime;
	    uint32_t         old_length, name_size;
	    int              temp_fd;
	    char             dns_name[MAX_DNS_NAME];
	    IADDR           *temp_iaddr;
	    FD_INFO         *temp_fd_info, prev_fd_info;
	    
	    name_size    = 0;
	    old_length   = length_read;
	    READ_DATA(&temp_ext, msg, length_read, sizeof(temp_ext));
	    
	    temp_ext     = ntohl(temp_ext);
	    s.status     = GET_BITS_HIGH(temp_ext, 8, 32);
	    s.C          = GET_BITS_MIDDLE(temp_ext, 24, 1);
	    s.I          = GET_BITS_MIDDLE(temp_ext, 23, 1);
	    s.L          = GET_BITS_MIDDLE(temp_ext, 22, 1);
	    s.Q          = GET_BITS_MIDDLE(temp_ext, 21, 1);
	    s.A          = GET_BITS_MIDDLE(temp_ext, 20, 1);
	    s.P          = GET_BITS_MIDDLE(temp_ext, 19, 1);
	    s.reserved   = 0;
	    s.scope      = GET_BITS_MIDDLE(temp_ext, 10, 3);
	    s.prefix_len = GET_BITS_MIDDLE(temp_ext, 7, 7);
	  
	    if(s.C){
	      READ_DATA(& ip_addr, msg, length_read, sizeof(struct in6_addr));
	    }
	    
	    if(s.L){
	      /*
	       * L bit can only be set if the client found out that the combination
	       * of bits sent by server was illegal and it copied the ext sent by server
	       */
	      if(s.status != 2){
		printf("There can be no lifetimes in release message\n");
		length_read = old_length + tl.length;
		break;/* break out of switch */
	      }
	      /* Just consume the lifetimes */
	      READ_DATA(& preferred_lifetime, msg, length_read, sizeof(preferred_lifetime));
	      READ_DATA(& valid_lifetime, msg, length_read, sizeof(valid_lifetime));
	      //preferred_lifetime = ntohl(preferred_lifetime);
	      //valid_lifetime     = ntohl(valid_lifetime);
	    }
	    
	    if((length_read - old_length) < tl.length){
	      /* This means DNS name is present. Read it off */
	      name_size           = tl.length - (length_read - old_length);
	      READ_DATA(dns_name, msg, length_read, name_size);
	      dns_name[name_size] = '\0';
#ifdef DEBUG
	      printf("DNS name = %s in IP ADDRESS EXTENSION\n", dns_name);
#endif	      
	    }
	    
	    if((s.C && !s.I && !s.A && !s.P) &&
	       (s.status == 0 || s.status == 1 || s.status == 2)){
	      /*
	       * This means the client is releasing the address because
	       * (0) It has finished using it.
	       * (1) It has discovered using DAD that IP addr is already
	       *     being used by a different node.
	       * (2) IP address has an illegal combination of bit settings
	       */
#ifdef DEBUG
	      printf("Appl wants to release address %s\n",
		     inet_ntop(AF_INET6, &ip_addr, name, 127));
#endif
	      /* Delete the address from interface addr list if present */
	      temp_fd = delete_address_info(&ip_addr, appl->interface);

	      if(temp_fd != -1){
		FD_CLR(temp_fd, &temp_f);
		delete_address(&ip_addr, appl->interface);
	      }
	      else{
#ifdef DEBUG
		printf("Warning! The released address isn't already present.\n");
#endif
	      }
	    }
	    else if(s.C && !s.I && (s.A || s.P) && s.status == 0){
#ifdef DEBUG
	      printf("Appl wants to delete DNS records\n");
#endif
	      /* CLD need do nothing */
	    }
	    else{
#ifdef DEBUG
	      printf("Illegal combination of bits in release message from appl\n");
#endif
	    }

	  }
	  break;
	  
	default:
	  printf("Invalid extension type in release message. Ignoring...\n");
	  length_read += tl.length;
	  break;
	}/* switch */
      }/* for length_read < len */

      memcpy(appl->msg, &rel, sizeof(rel));
      memcpy(&appl->msg[sizeof(rel)], msg + start_of_ext, len - start_of_ext);
      appl->msg_len                = sizeof(rel) + len - start_of_ext;

      appl->request_pending        = 1;

      appl->when_sent = current_time;
      appl->timeout   = REP_MSG_TIMEOUT;
      add_in_ms(&current_time,
		appl->timeout,
		&appl->next_transmission
		);
      
      send_msg_for_appl(appl);
    }
    break;


  case DHCP_RECONFIGURE_REPLY:
    {

    }
    break;

  default:
    break;
  }
}
