typedef struct hardware            HARDWARE;
typedef struct interface           INTERFACE;
typedef struct lease               LEASE;
typedef LEASE                      IADDR;
typedef struct timeval             TIME;
typedef struct appl_fd_msg         APPL_FD_MSG;
typedef struct fd_info             FD_INFO;


# include "../lib/solicit_info.h"
/*
 * Please remember IADDR = LEASE
 */

struct lease{
  /*
   * prev       : pointer to previous lease structure in linked list
   * next       : pointer to next lease structure in linked list
   * ip_addr    : addressed which is leased or is going to be leased
   * server_addr: the server associated with the above address.
   * server_name: name of the above server.
   * when_leased: time at which the above address was leased.
   * expiry     : time of expiry of ip_addr
   * pid        : which process obtained this lease.
   *
   * Note: server_addr = "::" i.e. unspecified address means address is obtaines
   * via manual configuration or otherwise not through DHCP server.
   */
  struct lease      *prev;
  struct lease      *next;
  struct in6_addr    ip_addr;
  char               dns_name[MAX_DNS_NAME];
  struct in6_addr    server_addr;
  TIME               when_leased;
  uint32_t           preferred_lifetime;
  uint32_t           valid_lifetime;
  INTERFACE         *interface;
};

struct hardware{
  uint8_t hw_addr[6];
};

struct interface{
  /*
   * Note that solicit id is only a 9 bit number. Therefore,
   * a value greater than 2^9 indicates uninitialised id i.e.
   * no solicit has been done on this interface.
   */
  struct interface    *prev;
  struct interface    *next;
  HARDWARE             hardware_addr;              /* Hardware address of the interface     */
  IADDR               *addr_list_head;             /* Head of link list of IPv6 address     */
  struct in6_addr     *link_local_addr;            /* Link local address of the i/f         */
  struct in6_addr     *loopback_addr;              /* Loopback addr if any                  */
  char                 name[IFNAMSIZ];             /* Name of the interface                 */
  uint32_t             index;                      /* Index of the interface                */

  LEASE               *lease_list_head;            /* List of unaccepted but offered leases */
  uint16_t             solicit_id;                 /* Solicit id last used on this interface*/
  uint16_t             xact_id;                    /* Transaction id last used on this i/f  */
  SERVER_SOLICIT_INFO *sol_list_head;              /* List of servers obtained via solicit  */
  TIME                 last_solicit_time;          /* Time of transmission of last solicit  */
  TIME                 advertise_time;             /* Time of 1'st advertise of last solicit*/
  FD_INFO             *link_local_fd;              /* fd opened on link local addr          */
  FD_INFO             *ip_addr_fd_head;            /* List of fds on non link local addrs   */
  
};

struct fd_info{
  /*
   * Note: This fd_info holds the UDP sockets used for listening
   * on CLIENT_PORT. These fds will also be used for replying to
   * server.
   * Also note that there can only be one fd per address because
   * all these are on same port viz. CLIENT_PORT.
   */
  FD_INFO             *next;
  int                  fd;                         /* UDP socket opened on CLIENT_PORT      */
  IADDR               *i_addr;                     /* Address to which this fd is bound     */
  struct sockaddr_in6  sockaddr;
};
