int try_cached_adv(FD_INIT_LIST *appl, INIT_DATA *data){
  /* 
   * Returns 0 if the cached data was old
   *         1 if success
   *         2 if some other problem
   */
  HEADER_CLIENTD_PACKET  header;
  TIME                   now;
  char                   padding[3];
  PREF_NO_OF_PREFIXES    pnp[MAX_SERVERS];
  SERVER_SOLICIT_INFO   *temp_serv_sol_info, **tsol;
  uint16_t               i, no_of_servers = 0, len = 0, padding_len;
  SUBNET_PREFIX_INFO    *temp_subnet_prefix_info;
  
  printf("try cached adv: 15\n");
  if(appl->interface->solicit_id < (1 << SOLICIT_ID_LEN) && data->c == 0){
    /*
     * Solicit id < 2^SOLICIT_ID_LEN means solicit has been done on this i/f
     * data->c == 0 means appl hasn't asked to release resources.
     * If these two conditions are met we can safely avoid sending
     * solicit msg if time of last solicit was not more earlier than
     * SOLICIT_INTERVAL * 1000 from current time.
     */
    
    printf("try cached adv: 25\n");
    if(gettimeofday(&now, NULL) != -1){
      printf("((((((((((((((((****))))))))))))))))if adv time = %u %u diff %u sol_int %u\n",
	     appl->interface->advertise_time.tv_sec, appl->interface->advertise_time.tv_usec,
	     DIFF_IN_MS(now, appl->interface->advertise_time), SOLICIT_INTERVAL * 1000);
      if(DIFF_IN_MS(now, appl->interface->advertise_time) < SOLICIT_INTERVAL * 1000){
	printf("try cached adv: 31\n");
	
#ifdef DEBUG
	printf("Using solicit info obtained earlier\n");
#endif
	
	/* Convert to network byte order */
	header.msg_type   = DHCP_ADVERTISE;
	header.error      = 0;
	header.pid        = htons(appl->pid);
	header.request_no = htonl(appl->request_no);
	len = sizeof(header);
	
	iov[0].iov_base   = (char *)&header;
	iov[0].iov_len    = len;
	i = 1;
	
	/* clear entire server solicit list as a new one would now result */
	clear_server_solicit_list(appl->sol_list_head);
	
	for(temp_serv_sol_info = appl->interface->sol_list_head, tsol = &appl->sol_list_head;
	    temp_serv_sol_info;
	    temp_serv_sol_info = temp_serv_sol_info->next, tsol = &(*tsol)->next){
	  
	  if(!(*tsol)){
	    *tsol = (SERVER_SOLICIT_INFO *) malloc(sizeof(SERVER_SOLICIT_INFO));
	    bzero(*tsol, sizeof(SERVER_SOLICIT_INFO));
	  }
	  
	  copy_server_solicit_info(*tsol, temp_serv_sol_info);

	  if(i > SIZE_OF_IOVEC - 3)
	    err_dump("Insufficient size of iovec\n");
	  /* Including relay address in reply */
	  iov[i].iov_base = (char *)& temp_serv_sol_info->relay_addr;
	  iov[i].iov_len  = sizeof(temp_serv_sol_info->relay_addr);
	  len            += iov[i].iov_len;
	  i++;
	  
	  /* Including server address in reply */
	  iov[i].iov_base = (char *)& temp_serv_sol_info->server_addr;
	  iov[i].iov_len  = sizeof(temp_serv_sol_info->server_addr);
	  len            += iov[i].iov_len;
	  i++;
	  
	  /* Including preference and no of prefixes info in reply */
	  pnp[no_of_servers].reserved       = 0;
	  pnp[no_of_servers].no_of_prefixes = temp_serv_sol_info->no_of_prefixes;
	  pnp[no_of_servers].preference     = temp_serv_sol_info->preference;
	  
	  iov[i].iov_base = (char *)& pnp[no_of_servers];
	  iov[i].iov_len  = sizeof(pnp[no_of_servers++]);
	  len            += iov[i].iov_len;
	  i++;
	  
	  if(data->p == 1){
	    /* Including prefixes info in reply */
	    for(temp_subnet_prefix_info = temp_serv_sol_info->subnet_prefix_info_head;
		temp_subnet_prefix_info;
		temp_subnet_prefix_info = temp_subnet_prefix_info->next){

	      if(i >= SIZE_OF_IOVEC)
		err_dump("Insufficient size of iovec\n");

	      iov[i].iov_base = & temp_subnet_prefix_info->prefix_len_octets[0];
	      /* Length of this data must be 1(for prefix-len) + Ceiling of prefix_len/8 bec
		 its in bits */
	      iov[i].iov_len  = 1 + (temp_subnet_prefix_info->prefix_len_octets[0] + 7)/8;
	      len            += iov[i].iov_len;
	      i++;
	    }
	    
	    /* Now we have to add padding if needed */
	    if((padding_len = (len % SIZE_OF_LONG)) != 0){

	      if(i >= SIZE_OF_IOVEC)
		err_dump("Insufficient size of iovec\n");

	      iov[i].iov_base = padding;
	      iov[i].iov_len  = SIZE_OF_LONG - padding_len;
	      len            += iov[i].iov_len;
	      i++;
	    }
	  }
	  
	  /* If only one reply was asked for break out of loop */
	  if(data->a == 0)
	    break;
	}

	appl->adv_received    = 1;
	appl->request_pending = 0;

	if(writev(appl->tcp_fd, &iov[0], i) != len){
	  printf("Error in sending reply to appl with pid = %u\n", appl->pid);
	  return SEND_ERROR;
	}

	return SUCCESS;
      }
      return OLD_CACHED_DATA;
    }

    err_dump("Error in gettimeofday. Exiting...");
    /* Not Reached */
  }
  return SOLICIT_NOT_DONE_OR_C_BIT_SET;
}
