/* Get interface name/index via sysctl */
# define DEBUG

#include <sys/types.h>
#include <stdio.h>
#include <errno.h>

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/sysctl.h>

#include <net/if.h>
#include <net/if_dl.h>
#ifndef if_type
#include <net/if_var.h>
#endif
#include <net/route.h>

#include <netinet/in.h>
#include <netinet/in6_var.h>

#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#ifdef DEBUG
#define static
#endif

#define ROUNDUP(a) \
        ((a) > 0 ? (1 + (((a) - 1) | (sizeof(long) - 1))) : sizeof(long))
#define ADVANCE(x, n) (x += ROUNDUP((n)->sdl_len))

static int mib[] = { CTL_NET, PF_ROUTE, 0, AF_MAX, NET_RT_IFLIST, 0 };

static void *getifs(char *, unsigned int *);

unsigned int
if_nametoindex(ifname)
	const char *ifname;
{
	unsigned int ret;
	
	if ((ifname == NULL) || (*ifname == '\0'))
		return(0);
	if (getifs((char *)ifname, &ret)) {
#ifdef DEBUG
		printf("if_nametoindex(%s)->%u\n",
		       ifname, ret);
#endif
		return(ret);
	}
#ifdef DEBUG
	printf("if_nametoindex(%s)->0!\n", ifname);
#endif
	return(0);
}

char *
if_indextoname(ifindex, ifname)
	unsigned int ifindex;
	char *ifname;
{
	bzero(ifname, IFNAMSIZ);
	if (getifs(ifname, &ifindex)) {
#ifdef DEBUG
		printf("if_indextoname(%d)->%s\n", ifindex, ifname);
#endif
		return(ifname);
	}
#ifdef DEBUG
	printf("if_indextoname(%d)->NULL\n", ifindex);
#endif
	return(NULL);
}

struct if_nameindex *
if_nameindex()
{
	struct if_nameindex *ret;

	ret = (struct if_nameindex *)getifs(NULL, NULL);
#ifdef DEBUG
	printf("if_nameindex->%p\n", ret);
#endif
	return(ret);
}

void
if_freenameindex(ptr)
	struct if_nameindex *ptr;
{
	free(ptr);
}

static void *
getifs(name, index)
char *name;
unsigned int *index;
{
	struct if_msghdr *ifm;
	struct if_nameindex *ifni;
	struct sockaddr_dl *sdl;
	caddr_t cp, rcp;
	int rsize = 0, size, rcount = 0;
	caddr_t buf, rbuf;

	if (sysctl(mib, 6, NULL, &size, NULL, 0) < 0) {
#ifdef DEBUG
		perror("sysctl0");
#endif
		return(0);
	}
#ifdef DEBUG
	printf("size = %d\n", size);
#endif
	if ((buf = malloc(size)) == NULL) {
#ifdef DEBUG
		printf( "malloc failed\n");
#endif
		return(0);
	}
	if (sysctl(mib, 6, buf, &size, NULL, 0) < 0) {
#ifdef DEBUG
		perror("sysctl");
#endif
		free(buf);
		return(0);
	}

	if (name)
		rsize = strlen(name);

	for (cp = buf; cp < buf + size; cp += ifm->ifm_msglen) {
		ifm = (struct if_msghdr *)cp;

		if (ifm->ifm_type != RTM_IFINFO)
			continue;
#ifdef DEBUG
		printf("msg_len  = %u\n",ifm->ifm_msglen);
		printf("version  = %u\n",ifm->ifm_version);
		printf("msg type = %u\n",ifm->ifm_type);
		printf("addrs    = %u\n",ifm->ifm_addrs);
		printf("index    = %u\n", ifm->ifm_index);
		printf("address  = %x\n", ifm->ifm_addrs);
		printf("ifi type = %u\n", ifm->ifm_data.ifi_type);
		printf("ifi phy  = %u\n", ifm->ifm_data.ifi_physical);
		printf("media addrlen = %u\n", ifm->ifm_data.ifi_addrlen);
		printf("media hdrlen  = %u\n", ifm->ifm_data.ifi_hdrlen);
		printf("mtu      = %u\n", ifm->ifm_data.ifi_mtu);
		printf("baudrate = %u\n", ifm->ifm_data.ifi_baudrate);
#endif
		if (ifm->ifm_addrs != (1 << RTAX_IFP)) {
#ifdef DEBUG
			printf( "bad ifm_addrs\n");
#endif
			continue;
		}

		sdl = (struct sockaddr_dl *)(ifm + 1);
		if (sdl->sdl_family != AF_LINK) {
#ifdef DEBUG
			printf("bad family: %d?: ", sdl->sdl_family);
#endif
			continue;
		}
			
#ifdef DEBUG
		printf("LL#%d/%d: %*s\n",
		       sdl->sdl_index,
		       sdl->sdl_type,
		       sdl->sdl_nlen,
		       sdl->sdl_data);
#endif
		if (sdl->sdl_nlen >= IFNAMSIZ) {
#ifdef DEBUG
			printf( "ifnamelen too large(%d)\n", sdl->sdl_nlen);
#endif
			continue;
		}
		if (name) {
			if (rsize &&
			    (rsize == (int)sdl->sdl_nlen) &&
			    (bcmp(sdl->sdl_data, name, rsize) == 0)) {
				*index = sdl->sdl_index;
				free(buf);
				return(name);
			}
			if (rsize == 0 &&
			    (*index == (unsigned int)sdl->sdl_index)) {
				strncpy(name, sdl->sdl_data, sdl->sdl_nlen);
				free(buf);
				return (name);
			}
		} else {
			rsize += sizeof(struct if_nameindex) + IFNAMSIZ;
			rcount++;
		}
#ifdef DEBUG
		rcp = (caddr_t)sdl;
		ADVANCE(rcp, sdl);

		if (rcp != (caddr_t)ifm + ifm->ifm_msglen)
			printf("delta = %d\n", rcp - ((caddr_t)ifm + ifm->ifm_msglen));
		printf("\n");
#endif
	}
	if (name || rsize == 0) {
		free(buf);
		return (0);
	}
	rsize += sizeof(struct if_nameindex);
	if ((rbuf = malloc(rsize)) == NULL) {
#ifdef DEBUG
		printf( "malloc failed\n");
#endif
		free(buf);
		return(0);
	}
	bzero(rbuf, rsize);
	ifni = (struct if_nameindex *)rbuf;
	rcp = (caddr_t)&ifni[rcount + 1];
	for (cp = buf; cp < buf + size; cp += ifm->ifm_msglen) {
		ifm = (struct if_msghdr *)cp;
		if (ifm->ifm_type != RTM_IFINFO)
			continue;
		if (ifm->ifm_addrs != (1 << RTAX_IFP))
			continue;

		sdl = (struct sockaddr_dl *)(ifm + 1);
		if (sdl->sdl_family != AF_LINK)
			continue;
		if (sdl->sdl_nlen >= IFNAMSIZ)
			continue;

		ifni->if_index = (unsigned int)sdl->sdl_index;
		ifni->if_name = rcp;
		strncpy(rcp, sdl->sdl_data, (size_t)sdl->sdl_nlen);

		ifni++;
		rcp += IFNAMSIZ;
	}
	free(buf);
	return(rbuf);
}


#ifdef DEBUG
void print_ifni(struct if_nameindex *);
int main(int, char **);

void
print_ifni(ifni)
	struct if_nameindex *ifni;
{
	if (ifni == NULL) {
		printf( "NULL?\n");
		return;
	}
	if ((ifni->if_index == 0) && (ifni->if_name == NULL))
		return;
	printf("%s : %u\n", ifni->if_name, ifni->if_index);
	print_ifni(++ifni);
}

int
main(argc, argv)
	int argc;
	char *argv[];
{
	char buf[IFNAMSIZ];


	getifs(NULL, NULL);
	printf("\n\nPress enter: ");
	getchar();
	if (argc == 1) {
		print_ifni(if_nameindex());
		exit(0);
	} else if (argc != 2) {
		printf( "bad argument count (%d)\n", argc);
		exit(1);
	}
	if (isdigit(*argv[1]))
		(void)if_indextoname(atoi(argv[1]), buf);
	else
		(void)if_nametoindex(argv[1]);
	exit(0);
}
#endif

int
main_interface(p)
	int *p;
{
	int mib[4] = { CTL_NET, PF_INET6, 0, IP6CTL_MAININTF };
	int midx, size = sizeof(midx);

	if (sysctl(mib, 4, &midx, &size, p, p ? sizeof(*p) : 0) < 0) {
#ifdef DEBUG
		perror("main_interface:  sysctl");
#endif
		return(-1);
	}
	if (size != sizeof(midx)) {
#ifdef DEBUG
		printf( "main_interface: bad size (%d)\n", size);
#endif
		return(-1);
	}
	if (p)
		*p = midx;
	return(midx);
}

int
main_site(p)
	int *p;
{
	int mib[4] = { CTL_NET, PF_INET6, 0, IP6CTL_MAINSITE };
	int msite, size = sizeof(msite);

	if (sysctl(mib, 4, &msite, &size, p, sizeof(*p)) < 0) {
#ifdef DEBUG
		perror("main_site:  sysctl");
#endif
		return(-1);
	}
	if (size != sizeof(msite)) {
#ifdef DEBUG
		printf( "main_site: bad size (%d)\n", size);
#endif
		return(-1);
	}
	if (p)
		*p = msite;
	return(msite);
}

