/* Get all interface infos via sysctl */

#include <sys/types.h>
#include <sys/time.h>
#include <stdio.h>
#include <errno.h>
#include <signal.h>

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/sysctl.h>

#include <net/if.h>
#include <net/if_dl.h>
#include <net/route.h>

#include <netinet/in.h>

#include <netdb.h>
#include <arpa/inet.h>

#include <string.h>
#include <stdlib.h>
#include <syslog.h>

int mib[] = { CTL_NET, PF_ROUTE, 0, 0, NET_RT_IFLIST, 0 };
int size;
caddr_t buf;
struct if_msghdr *ifm;
struct ifa_msghdr *ifam;
struct sockaddr *sa;
struct timeval tv;

main() {
	caddr_t cp;

	if (sysctl(mib, 6, NULL, &size, NULL, 0) < 0) {
		perror("sysctl0");
		exit(1);
	}
	printf("size = %d\n", size);
	if ((buf = malloc(size)) == NULL) {
		fprintf(stderr, "malloc failed\n");
		exit(1);
	}
	if (sysctl(mib, 6, buf, &size, NULL, 0) < 0) {
		perror("sysctl");
		exit(1);
	}
	if (gettimeofday(&tv, NULL) < 0) {
		perror("gettimeofday");
		exit(1);
	}

	for (cp = buf; cp < buf + size; cp += ifm->ifm_msglen) {
		ifm = (struct if_msghdr *)cp;
		ifam = (struct ifa_msghdr *)cp;

		if (ifm->ifm_type == RTM_IFINFO)
			ifinfo();
		else
			ifaddr();
	}
}

#define ROUNDUP(a) \
        ((a) > 0 ? (1 + (((a) - 1) | (sizeof(long) - 1))) : sizeof(long))
#define ADVANCE(x, n) (x += ROUNDUP((n)->sa_len))

ifinfo() {
	caddr_t cp;
	int len, i;

	cp = (caddr_t)(ifm + 1);
	len = ifm->ifm_msglen;

	printf("Interface Info #%d\n", ifm->ifm_index);

	printf("\tflags = %x\n", (u_int)ifm->ifm_flags);
	printf("\ttype = %d\n", ifm->ifm_data.ifi_type);
	printf("\taddrlen = %d\n", ifm->ifm_data.ifi_addrlen);

	printf("\thdrlen = %d\n", ifm->ifm_data.ifi_hdrlen);
	printf("\tmtu = %d\n", ifm->ifm_data.ifi_mtu);
	printf("\tmetric = %d\n", ifm->ifm_data.ifi_metric);
	printf("\tbaudrate = %d\n", ifm->ifm_data.ifi_baudrate);
	printf("\tipackets = %d\n", ifm->ifm_data.ifi_ipackets);
	printf("\tierrors = %d\n", ifm->ifm_data.ifi_ierrors);
	printf("\topackets = %d\n", ifm->ifm_data.ifi_opackets);
	printf("\toerrors = %d\n", ifm->ifm_data.ifi_oerrors);
	printf("\tcollisions = %d\n", ifm->ifm_data.ifi_collisions);
	printf("\tibytes = %d\n", ifm->ifm_data.ifi_ibytes);
	printf("\tobytes = %d\n", ifm->ifm_data.ifi_obytes);
	printf("\timcasts = %d\n", ifm->ifm_data.ifi_imcasts);
	printf("\tomcasts = %d\n", ifm->ifm_data.ifi_omcasts);
	printf("\tiqdrops = %d\n", ifm->ifm_data.ifi_iqdrops);
	printf("\tnoproto = %d\n", ifm->ifm_data.ifi_noproto);
	tv.tv_sec -= ifm->ifm_data.ifi_lastchange.tv_sec;
	tv.tv_usec -= ifm->ifm_data.ifi_lastchange.tv_usec;
	if (tv.tv_usec < 0) {
		tv.tv_usec += 1000000;
		tv.tv_sec -= 1;
	}
	printf("\tlastchange = %g s ago\n",
	       (double)(tv.tv_sec + (tv.tv_usec / 1000000.)));

	printf("\t%x addresses\n", ifm->ifm_addrs);

	for (i = 0; i < RTAX_MAX; i++)
		if (ifm->ifm_addrs & (1 << i)) {
			sa = (struct sockaddr *)cp;
			bit(i);
			sock();
			ADVANCE(cp, sa);
		}

	if (cp != (caddr_t)ifm + len)
		printf("delta = %d\n", cp - ((caddr_t)ifm + len));
	printf("\n");
}

ifaddr() {
	caddr_t cp;
	int len, i;

	cp = (caddr_t)(ifam + 1);
	len = ifam->ifam_msglen;

	printf("Interface Address #%d\n", ifam->ifam_index);

	printf("\tflags = %x\n", ifam->ifam_flags);
	printf("\tmetric = %x\n", ifam->ifam_metric);

	printf("\t%x addresses\n", ifam->ifam_addrs);

	for (i = 0; i < RTAX_MAX; i++)
		if (ifam->ifam_addrs & (1 << i)) {
			sa = (struct sockaddr *)cp;
			bit(i);
			sock();
			ADVANCE(cp, sa);
		}

	if (cp != (caddr_t)ifam + len)
		printf("delta = %d\n", cp - ((caddr_t)ifam + len));
	printf("\n");
}

bit(i)
int i;
{
	printf("\t\t");
	switch (i) {
	    case RTAX_DST:
		printf("DST: ");
		return;

	    case RTAX_GATEWAY:
		printf("GATEWAY: ");
		return;

	    case RTAX_NETMASK:
		printf("NETMASK: ");
		return;

	    case RTAX_GENMASK:
		printf("GENMASK: ");
		return;

	    case RTAX_IFP:
		printf("IFP: ");
		return;

	    case RTAX_IFA:
		printf("IFA: ");
		return;

	    case RTAX_AUTHOR:
		printf("AUTHOR: ");
		return;

	    case RTAX_BRD:
		printf("BRD: ");
		return;

	    default:
		printf("??: ");
		return;
	}
}

sock() {
	u_char *cp;
	char buf[64];
	int i;

	switch (sa->sa_family) {
	    case AF_UNSPEC:
		printf("mask %d\n", sa->sa_len);
		break;

	    case AF_INET:
#define SIN(x)	((struct sockaddr_in *)x)
		printf("IP: %s\n", inet_ntoa(SIN(sa)->sin_addr));
		return;

	    case AF_INET6:
#define SIN6(x)	((struct sockaddr_in6 *)x)
		printf("IPv6: %s\n",
		       inet_ntop(AF_INET6, &SIN6(sa)->sin6_addr,
				 buf, sizeof(buf)));
		return;

	    case AF_LINK:
#define SDL(x)	((struct sockaddr_dl *)x)
	      	printf("LL#%d/%d: %*s:",
		       SDL(sa)->sdl_index,
		       SDL(sa)->sdl_type,
		       SDL(sa)->sdl_nlen,
		       SDL(sa)->sdl_data);
		cp = (u_char *)(SDL(sa)->sdl_data + SDL(sa)->sdl_nlen);
		for (i = SDL(sa)->sdl_alen; i > 0; i--)
			printf("%02x", *cp++);
		if (SDL(sa)->sdl_slen) {
			printf("/");
			for (i = SDL(sa)->sdl_slen; i > 0; i--)
				printf("%02x", *cp++);
		}
		printf("\n");
		return;

	    default:
		printf("%d?: ", sa->sa_family);
		cp = (u_char *)sa;
		for (i = sa->sa_len; i > 0; i--)
			printf("%02x ", *cp++);
		printf("\n");
	}
}
