/* Get all interface infos via sysctl */

#include <sys/types.h>
#include <sys/time.h>
#include <stdio.h>
#include <errno.h>
#include <signal.h>

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/sysctl.h>

#include <net/if.h>
#include <net/if_dl.h>
#include <net/route.h>

#include <netinet/in.h>

#include <netdb.h>
#include <arpa/inet.h>

#include <string.h>
#include <stdlib.h>
#include <syslog.h>

#define SERVER_NAME_SIZE 128

typedef struct hardware        HARDWARE;
typedef struct interface       INTERFACE;
//typedef struct interface_state INTERFACE_STATE;
typedef struct lease           LEASE;
typedef LEASE                  IADDR;
typedef time_t                 TIME;

struct lease{//IADDR = LEASE
  struct lease   *prev;
  struct lease   *next;
  struct in6_addr ip_addr;
  struct in6_addr server_addr;   // 0 i.e. UNSPECIFIED ADDRESS: statically obtained not thro' DHCP.
  char            server_name[SERVER_NAME_SIZE];
  TIME            when_leased;
  TIME            expiry;
};

struct hardware{
  uint8_t hw_addr[6];
};

struct interface{
  struct interface *prev;
  struct interface *next;
  HARDWARE          hardware_addr;
  IADDR            *addr_list_head;
  struct in6_addr  *link_local_addr;
  struct in6_addr  *loopback_addr;
  char              name[IFNAMSIZ];
  uint              index;
  LEASE            *lease_list_head;
  
};

void initialise_addr(IADDR *addr){
  inet_ntop(AF_INET6, &addr->ip_addr, "::", 3);
  inet_ntop(AF_INET6, &addr->server_addr, "::", 3);
  addr->server_name[0] = '\0';
  addr->when_leased    = 0;
  addr->expiry         = 0;
  addr->next           = addr->prev = NULL;
}

INTERFACE *find_interface(uint index, INTERFACE *head){
  INTERFACE *tmp = head;

  while(tmp != NULL){
    if(tmp->index == index)
      break;
    tmp = tmp->next;
  }
  return tmp;
}

void insert_in_list(IADDR *addr, IADDR **head){
                // (INTREFACE *addr, INTERFACE **head){
  if(addr == NULL){

#ifdef DEBUG
    printf("Nothing to insert into list\n");
#endif
    return;
  }

  if(*head != NULL){
    (*head)->prev = addr;
  }
  /*
  if(*head->next != NULL){
    *head->next->prev = addr;
    addr->next       = *head->next;
  }

  *head->next = addr;
  addr->prev = *head;
  */
  addr->next = *head;
  *head       = addr;
  (*head)->prev = NULL;
}

void delete_from_list(IADDR *addr, IADDR **head){

  if(addr == NULL){

#ifdef DEBUG
    printf("Nothing to delete from list\n");
#endif
    return;
  }

  if(*head == NULL){

#ifdef DEBUG
    printf("Can't delete from empty list\n");
#endif
    return;
  }

  if(addr->prev != NULL)
    addr->prev->next = addr->next;

  if(addr->next != NULL)
    addr->next->prev = addr->prev;

  if(addr == *head)
    *head = (*head)->next;

  // free(addr);
}

void print(INTERFACE *head){
  IADDR     *tmp_addr;
  INTERFACE *tmp_interface = head;
  char buf[128];
  uint8_t i;

  while(tmp_interface != NULL){
    if(tmp_interface->link_local_addr != NULL){
      printf("Link local addr of [%u] name <%s> = %s Link layer address = ",
	     tmp_interface->index,
	     tmp_interface->name,
	     inet_ntop(AF_INET6, tmp_interface->link_local_addr, buf, sizeof(buf)));

      for(i = 0; i < 6; i++)
	printf("%02x",tmp_interface->hardware_addr.hw_addr[i]);
      printf("\n");

      tmp_addr = tmp_interface->addr_list_head;
      printf("\t\tIPv6 addr list ");
      while(tmp_addr != NULL){
	printf("%s ",inet_ntop(AF_INET6, & tmp_addr->ip_addr, buf, sizeof(buf)));
	tmp_addr = tmp_addr->next;
      }
      printf("\n");
    }
    tmp_interface = tmp_interface->next;
  }
}

# define ADDRESS_FAMILY AF_INET6

int                mib[] = { CTL_NET, PF_ROUTE, 0, ADDRESS_FAMILY, NET_RT_IFLIST, 0 };
int                size;
caddr_t            buf;
struct if_msghdr  *ifm;
struct ifa_msghdr *ifam;
struct sockaddr   *sa;
struct timeval     tv;
INTERFACE         *in_list_head, *temp;
IADDR              addr;

main() {
	caddr_t cp;
	int interface_in_list;
	/*
	struct in6_addr ip_addr;
	struct in6_addr server_addr;  
	char            server_name[SERVER_NAME_SIZE];
	TIME            when_leased;
	TIME            expiry;
	struct lease   *prev;
	struct lease   *next;
	*/
	/*
	memset((char *) &addr, (unsigned char)0, sizeof(addr));
	printf("ip_addr unspecified = %d\n",IN6_IS_ADDR_UNSPECIFIED(& addr.ip_addr));
	printf("server_addr unspecified = %d\n",IN6_IS_ADDR_UNSPECIFIED(& addr.server_addr));
	printf("name = <%s> when_leased = %u expiry = %d\n",addr.server_name);
	if(addr.server_name[0] == '\0')
	  printf("NAME NULL");
	if(addr.prev == NULL)
	  printf("prev NULL ");
	if(addr.next == NULL)
	printf("next NULL \n");*/

	if (sysctl(mib, 6, NULL, &size, NULL, 0) < 0) {
		perror("sysctl0");
		exit(1);
	}
	//printf("size = %d\n", size);
	if ((buf = malloc(size)) == NULL) {
		fprintf(stderr, "malloc failed\n");
		exit(1);
	}
	if (sysctl(mib, 6, buf, &size, NULL, 0) < 0) {
		perror("sysctl");
		exit(1);
	}
	if (gettimeofday(&tv, NULL) < 0) {
		perror("gettimeofday");
		exit(1);
	}
	/*
	printf("LEASE = %u H/W = %u I/F = %u TIME = %u\n",
	       sizeof(LEASE),
	       sizeof(HARDWARE),
	       sizeof(INTERFACE),       
	       sizeof(TIME));
	*/
	for (cp = buf; cp < buf + size; cp += ifm->ifm_msglen) {
		ifm = (struct if_msghdr *)cp;
		ifam = (struct ifa_msghdr *)cp;
		interface_in_list = 1;
		temp = find_interface(ifm->ifm_index, in_list_head);
		if(temp == NULL){
		  //printf("Not found interface %u \n",ifm->ifm_index);
		  temp = (INTERFACE *) malloc(sizeof(INTERFACE));
		  memset(temp, 0, sizeof(*temp));
		  temp->index = ifm->ifm_index;
		  interface_in_list = 0;
		}

		if (ifm->ifm_type == RTM_IFINFO)
			ifinfo();
		else
			ifaddr();
		/*
		if(temp->link_local_addr == NULL){
		  free(temp);
		  }*/
		//else{
		if(interface_in_list == 0){
		  //printf("inserting in list index = %u\n",ifm->ifm_index);
		  insert_in_list((IADDR *)temp, (IADDR **)&in_list_head);
		  if(in_list_head == NULL)
		    printf("Null after inserting");
		  //getchar();
		}
		  //}
	}
	print(in_list_head);
}

#define ROUNDUP(a) \
        ((a) > 0 ? (1 + (((a) - 1) | (sizeof(long) - 1))) : sizeof(long))
#define ADVANCE(x, n) (x += ROUNDUP((n)->sa_len))

ifinfo() {
	caddr_t cp;
	int len, i;

	cp = (caddr_t)(ifm + 1);
	len = ifm->ifm_msglen;

	//temp->index = ifm->ifm_index;
	/*
	printf("Interface Info for interface %d\n", ifm->ifm_index);

	printf("\tflags = %x\n", (u_int)ifm->ifm_flags);
	printf("\ttype = %d\n", ifm->ifm_data.ifi_type);
	printf("\taddrlen = %d\n", ifm->ifm_data.ifi_addrlen);

	printf("\thdrlen = %d\n", ifm->ifm_data.ifi_hdrlen);
	printf("\tmtu = %d\n", ifm->ifm_data.ifi_mtu);
	printf("\tmetric = %d\n", ifm->ifm_data.ifi_metric);
	printf("\tbaudrate = %d\n", ifm->ifm_data.ifi_baudrate);
	printf("\tipackets = %d\n", ifm->ifm_data.ifi_ipackets);
	printf("\tierrors = %d\n", ifm->ifm_data.ifi_ierrors);
	printf("\topackets = %d\n", ifm->ifm_data.ifi_opackets);
	printf("\toerrors = %d\n", ifm->ifm_data.ifi_oerrors);
	printf("\tcollisions = %d\n", ifm->ifm_data.ifi_collisions);
	printf("\tibytes = %d\n", ifm->ifm_data.ifi_ibytes);
	printf("\tobytes = %d\n", ifm->ifm_data.ifi_obytes);
	printf("\timcasts = %d\n", ifm->ifm_data.ifi_imcasts);
	printf("\tomcasts = %d\n", ifm->ifm_data.ifi_omcasts);
	printf("\tiqdrops = %d\n", ifm->ifm_data.ifi_iqdrops);
	printf("\tnoproto = %d\n", ifm->ifm_data.ifi_noproto);
	
	tv.tv_sec -= ifm->ifm_data.ifi_lastchange.tv_sec;
	tv.tv_usec -= ifm->ifm_data.ifi_lastchange.tv_usec;
	if (tv.tv_usec < 0) {
		tv.tv_usec += 1000000;
		tv.tv_sec -= 1;
	}
	
	printf("\tlastchange = %g s ago\n",
	       (double)(tv.tv_sec + (tv.tv_usec / 1000000.)));

	printf("\t%x addresses\n", ifm->ifm_addrs);
	*/
	for (i = 0; i < RTAX_MAX; i++)
		if (ifm->ifm_addrs & (1 << i)) {
			sa = (struct sockaddr *)cp;
			bit(i);
			sock();
			ADVANCE(cp, sa);
		}

	if (cp != (caddr_t)ifm + len)
		printf("delta = %d\n", cp - ((caddr_t)ifm + len));
	//printf("\nreturnning from ifinfo\n\n");
}

ifaddr() {
	caddr_t cp;
	int len, i;

	cp = (caddr_t)(ifam + 1);
	len = ifam->ifam_msglen;
	/*
	printf("Interface Address #%d\n", ifam->ifam_index);

	printf("\tflags = %x\n", ifam->ifam_flags);
	printf("\tmetric = %x\n", ifam->ifam_metric);

	printf("\t%x addresses\n", ifam->ifam_addrs);
	*/
	for (i = 0; i < RTAX_MAX; i++)
		if (ifam->ifam_addrs & (1 << i)) {
			sa = (struct sockaddr *)cp;
			bit(i);
			sock();
			ADVANCE(cp, sa);
		}

	if (cp != (caddr_t)ifam + len)
		printf("delta = %d\n", cp - ((caddr_t)ifam + len));
	//printf("\nreturning from ifaddr\n\n");
}

bit(i)
int i;
{
  //printf("\t\t");
	switch (i) {
	    case RTAX_DST:
	      //printf("DST: ");
		return;

	    case RTAX_GATEWAY:
	      //printf("GATEWAY: ");
		return;

	    case RTAX_NETMASK:
	      //printf("NETMASK: ");
		return;

	    case RTAX_GENMASK:
	      //printf("GENMASK: ");
		return;

	    case RTAX_IFP:
	      //printf("IFP: ");
		return;

	    case RTAX_IFA:
	      //printf("IFA: ");
		return;

	    case RTAX_AUTHOR:
	      //printf("AUTHOR: ");
		return;

	    case RTAX_BRD:
	      //printf("BRD: ");
		return;

	    default:
	      //printf("??: ");
		return;
	}
}

sock() {
	u_char *cp;
	char buf[64];
	int i;
	IADDR *temp_lease;

	switch (sa->sa_family) {
	    case AF_UNSPEC:
	      //printf("mask %d\n", sa->sa_len);
		break;

	    case AF_INET:
	      //#define SIN(x)	((struct sockaddr_in *)x)
	      //printf("IP: %s\n", inet_ntoa(SIN(sa)->sin_addr));
		return;

	    case AF_INET6:
#define SIN6(x)	((struct sockaddr_in6 *)x)
	      temp_lease = (IADDR *)malloc(sizeof(IADDR));
	      memset((char *) temp_lease, 0, sizeof(*temp_lease));
	      temp_lease->ip_addr = SIN6(sa)->sin6_addr;

	      if(IN6_IS_ADDR_LOOPBACK(& temp_lease->ip_addr)){
		temp->loopback_addr    = (struct in6_addr *)malloc(sizeof(struct in6_addr));
		*temp->loopback_addr   = temp_lease->ip_addr;
		free(temp_lease);

	      }
	      else{
		if(IN6_IS_ADDR_LINKLOCAL(& temp_lease->ip_addr)){
		  temp->link_local_addr  = (struct in6_addr *)malloc(sizeof(struct in6_addr));
		  *temp->link_local_addr = temp_lease->ip_addr;
		  free(temp_lease);
		}
		else
		  insert_in_list(temp_lease, & temp->addr_list_head);
	      }
	      /*
		printf("IPv6: %s\n",
		       inet_ntop(AF_INET6, &SIN6(sa)->sin6_addr,
		       buf, sizeof(buf)));*/
		return;

	    case AF_LINK:
#define SDL(x)	((struct sockaddr_dl *)x)
	      /*
	      printf("-------------LL<%d><%d>: <%*s>:",
		       SDL(sa)->sdl_index,
		       SDL(sa)->sdl_type,
		       SDL(sa)->sdl_nlen,
		       SDL(sa)->sdl_data);
	      */
	      sprintf(temp->name,"%*s",
		       SDL(sa)->sdl_nlen,
		       SDL(sa)->sdl_data);
	      	//printf("%*s", SDL(sa)->sdl_data);

		cp = (u_char *)(SDL(sa)->sdl_data + SDL(sa)->sdl_nlen);
		
		//for (i = SDL(sa)->sdl_alen; i > 0; i--)
		for(i = 0; i < SDL(sa)->sdl_alen; i++){
		  //sprintf( ((char *)&temp->hardware_addr + i), "%02x", *cp++);
		  temp->hardware_addr.hw_addr[i] = *cp++;
		  //printf("%02x", temp->hardware_addr.hw_addr[i]);
		}
		/*
		if (SDL(sa)->sdl_slen) {
			printf("/");
			for (i = SDL(sa)->sdl_slen; i > 0; i--)
				printf("%02x", *cp++);
				}*/
				//printf("\n");
		return;
	      
	    default:
	      /*printf("%d?: ", sa->sa_family);
		cp = (u_char *)sa;
		for (i = sa->sa_len; i > 0; i--)
			printf("%02x ", *cp++);
			printf("\n");*/
	}
}
